from mstrio.connection import Connection
from mstrio.utils.api_helpers import unpack_information
from mstrio.utils.error_handlers import ErrorHandler


@unpack_information
@ErrorHandler(err_msg="Error getting an incremental refresh report")
def get_incremental_refresh_report(
    connection: Connection,
    id: str,
    project_id: str | None = None,
    show_expression_as: list[str] | None = None,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = True,
):
    """Get a definition of incremental refresh report

    Args:
        connection: Strategy One REST API connection object
        id (str): Incremental Refresh Report's ID
        project_id (str, optional): ID of a project
        show_expression_as (list[str], optional): Specifies the format in
            which the expressions are returned in response
            Available values: 'tokens', 'tree'
            If omitted, the expression is returned in 'text' format
            If 'tree', the expression is returned in 'text' and 'tree' formats.
            If 'tokens', the expression is returned in 'text' and 'tokens'
            formats.
        show_filter_tokens (bool, optional): specify whether `qualification`
            is returned in `tokens` format, along with `text` and `tree`
            format
        show_advanced_properties (bool, optional): Specify whether to retrieve
            the values of the advanced properties.
            The advanced properties are presented in the following groups:
                "vldbProperties": A list of properties as determined by
                    the common infrastructure.
                "metricJoinTypes": A list of Metric Join Types, one for each
                    metric that appears in the template.
                "attributeJoinTypes": A list of Attribute Join Types, one for
                    each attribute that appears in the template.
            If omitted or false, nothing will be returned for the advanced
                properties.
            If true, all applicable advanced properties are returned.


    Return:
        HTTP response object. Expected status: 200
    """
    return connection.get(
        endpoint=f'/api/model/incrementalRefresh/{id}',
        headers={'X-MSTR-ProjectID': project_id} if project_id else None,
        params={
            'showExpressionAs': show_expression_as,
            'showFilterTokens': str(show_filter_tokens).lower(),
            'showAdvancedProperties': str(show_advanced_properties).lower(),
        },
    )


@unpack_information
@ErrorHandler(err_msg="Error creating an incremental refresh report")
def create_incremental_refresh_report(
    connection: Connection,
    body: dict,
    project_id: str | None = None,
    show_expression_as: list[str] | None = None,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = True,
):
    """Create a new incremental refresh report,
    based on the definition provided in request body.

    Args:
        connection: Strategy One REST API connection object
        body (dict): Incremental Refresh Report's creation data
        project_id (str, optional): ID of a project
        show_expression_as (list[str], optional): Specifies the format in
            which the expressions are returned in response
            Available values: 'tokens', 'tree'
            If omitted, the expression is returned in 'text' format
            If 'tree', the expression is returned in 'text' and 'tree' formats.
            If 'tokens', the expression is returned in 'text' and 'tokens'
            formats.
        show_filter_tokens (bool, optional): specify whether `qualification`
            is returned in `tokens` format, along with `text` and `tree`
            format
        show_advanced_properties (bool, optional): Specify whether to retrieve
            the values of the advanced properties.
            The advanced properties are presented in the following groups:
                "vldbProperties": A list of properties as determined by
                    the common infrastructure.
                "metricJoinTypes": A list of Metric Join Types, one for each
                    metric that appears in the template.
                "attributeJoinTypes": A list of Attribute Join Types, one for
                    each attribute that appears in the template.
            If omitted or false, nothing will be returned for the advanced
                properties.
            If true, all applicable advanced properties are returned.


    Return:
        HTTP response object. Expected status: 201
    """
    return connection.post(
        endpoint='/api/model/incrementalRefresh',
        headers={'X-MSTR-ProjectID': project_id} if project_id else None,
        params={
            'showExpressionAs': show_expression_as,
            'showFilterTokens': str(show_filter_tokens).lower(),
            'showAdvancedProperties': str(show_advanced_properties).lower(),
        },
        json=body,
    )


@unpack_information
@ErrorHandler(err_msg="Error updating an incremental refresh report")
def update_incremental_refresh_report(
    connection: Connection,
    id: str,
    body: dict,
    project_id: str | None = None,
    show_expression_as: list[str] | None = None,
    show_filter_tokens: bool = False,
    show_advanced_properties: bool = True,
):
    """Update a specified incremental refresh report,
    based on the definition provided in request body.

    Args:
        connection: Strategy One REST API connection object
        id (str): Incremental Refresh Report's ID
        body (dict): Incremental Refresh Report's update data
        project_id (str, optional): ID of a project
        show_expression_as (list[str], optional): Specifies the format in
            which the expressions are returned in response
            Available values: 'tokens', 'tree'
            If omitted, the expression is returned in 'text' format
            If 'tree', the expression is returned in 'text' and 'tree' formats.
            If 'tokens', the expression is returned in 'text' and 'tokens'
            formats.
        show_filter_tokens (bool, optional): specify whether `qualification`
            is returned in `tokens` format, along with `text` and `tree`
            format
        show_advanced_properties (bool, optional): Specify whether to retrieve
            the values of the advanced properties.
            The advanced properties are presented in the following groups:
                "vldbProperties": A list of properties as determined by
                    the common infrastructure.
                "metricJoinTypes": A list of Metric Join Types, one for each
                    metric that appears in the template.
                "attributeJoinTypes": A list of Attribute Join Types, one for
                    each attribute that appears in the template.
            If omitted or false, nothing will be returned for the advanced
                properties.
            If true, all applicable advanced properties are returned.

    Return:
        HTTP response object. Expected status: 200
    """
    return connection.put(
        endpoint=f'/api/model/incrementalRefresh/{id}',
        headers={'X-MSTR-ProjectID': project_id} if project_id else None,
        params={
            'showExpressionAs': show_expression_as,
            'showFilterTokens': str(show_filter_tokens).lower(),
            'showAdvancedProperties': str(show_advanced_properties).lower(),
        },
        json=body,
    )


@unpack_information
@ErrorHandler(err_msg="Error getting an incremental refresh report's vldb properties")
def get_incremental_refresh_report_vldb_properties(
    connection: Connection, id: str, project_id: str | None = None
):
    """Get vldb properties of an incremental refresh report.

    Args:
        connection: Strategy One REST API connection object
        id (str): Incremental Refresh Report's ID
        project_id (str): ID of a project

    Return:
        HTTP response object. Expected status: 200
    """
    return connection.get(
        endpoint=f'/api/model/incrementalRefresh/{id}/applicableVldbProperties',
        headers={'X-MSTR-ProjectID': project_id} if project_id else None,
    )


@ErrorHandler(err_msg="Error executing an incremental refresh report")
def execute_incremental_refresh_report(
    connection: Connection, id: str, project_id: str, fields: str | None = None
):
    """Execute an incremental refresh report.

    Args:
        connection: Strategy One REST API connection object
        id (str): Incremental Refresh Report's ID
        project_id (str): ID of a project
        fields: A whitelist of top-level fields separated by commas.
            Allow the client to selectively retrieve fields in the response.

    Return:
        HTTP response object. Expected status: 202
    """
    return connection.post(
        endpoint=f'/api/incrementalRefresh/{id}',
        headers={'X-MSTR-ProjectID': project_id},
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error requesting preview data for an incremental refresh report")
def request_incremental_refresh_report_preview_data(
    connection: Connection,
    id: str,
    instance_id: str,
    project_id: str,
    fields: str | None = None,
):
    """Request preview data for an incremental refresh report.

    Args:
        connection: Strategy One REST API connection object
        id (str): Incremental Refresh Report's ID
        instance_id (str): Instance's ID of an Incremental Refresh Report
        project_id (str): ID of a project
        fields: A whitelist of top-level fields separated by commas.
            Allow the client to selectively retrieve fields in the response.

    Return:
        HTTP response object. Expected status: 202
    """
    return connection.post(
        endpoint=f'/api/incrementalRefresh/{id}/instances/{instance_id}/data',
        headers={'X-MSTR-ProjectID': project_id},
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error getting preview data for an incremental refresh report")
def get_incremental_refresh_report_preview_data(
    connection: Connection,
    id: str,
    instance_id: str,
    project_id: str,
    offset: int | None = None,
    limit: int | None = None,
    fields: str | None = None,
):
    """Get preview data for an incremental refresh report.

    Args:
        connection: Strategy One REST API connection object
        id (str): Incremental Refresh Report's ID
        instance_id (str): Instance's ID of an Incremental Refresh Report
        project_id (str): ID of a project
        offset (int):
        limit (int):
        fields: A whitelist of top-level fields separated by commas.
            Allow the client to selectively retrieve fields in the response.

    Return:
        HTTP response object. Expected status: 200
    """
    return connection.get(
        endpoint=f'/api/incrementalRefresh/{id}/instances/{instance_id}/data',
        headers={'X-MSTR-ProjectID': project_id},
        params={
            'offset': offset,
            'limit': limit,
            'fields': fields,
        },
    )


@ErrorHandler(err_msg="Error creating instance of the incremental refresh report")
def create_incremental_refresh_report_instance(connection: Connection, id: str):
    """Create instance of an incremental refresh report.

    Args:
        connection: Strategy One REST API connection object
        id (str): Incremental Refresh Report's ID

        Return:
            HTTP response object. Expected status: 201
    """
    return connection.post(
        endpoint=f'/api/model/incrementalRefresh/{id}/instances',
    )


@ErrorHandler(err_msg="Error deleting instance of the incremental refresh report")
def delete_incremental_refresh_report_instance(
    connection: Connection, id: str, instance_id: str
):
    """Delete instance of an incremental refresh report.

    Args:
        connection: Strategy One REST API connection object
        id (str): Incremental Refresh Report's ID
        instance_id (str): Incremental Refresh Report Instance ID

        Return:
            HTTP response object. Expected status: 201
    """
    return connection.delete(
        endpoint=f'/api/model/incrementalRefresh/{id}/instances',
        headers={
            'X-MSTR-MS-INSTANCE': instance_id,
        },
    )



"""This is the sample script to show how to manage filters.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

from mstrio.connection import Connection, get_connection
from mstrio.modeling import (
    ExpressionFormat, Filter, list_filters, SchemaManagement, SchemaUpdateType
)

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn: Connection = get_connection(workstationData, PROJECT_NAME)

# Example filter qualification data with expression specified as tree
FILTER_QUALIFICATION_AS_TREE = {
    "tree": {
        "type": "predicate_metric_qualification",
        "predicateTree": {
            "function": "greater",
            "parameters": [
                {
                    "parameterType": "constant",
                    "constant": {
                        "type": "int64", "value": "1"
                    },
                }
            ],
            "levelType": "none",
            "metric": {
                "objectId": "CE8DAE7048998FFECF77418BDF2343B8",
                "subType": "metric",
                "name": "Item Count",
            },
            "metricFunction": "value",
            "isIndependent": 0,
            "nullInclude": 0,
        },
    },
}

# Example filter qualification data with expression specified as tokens
FILTER_QUALIFICATION_AS_TOKENS = {
    "tokens": [
        {
            "level": "resolved",
            "state": "initial",
            "value": "%",
            "type": "character",
        },
        {
            "level": "resolved",
            "state": "initial",
            "value": "[Item Count]",
            "type": "object_reference",
        },
        {
            "level": "resolved",
            "state": "initial",
            "value": ">",
            "type": "character",
        },
        {
            "level": "resolved", "state": "initial", "value": "1", "type": "unknown"
        },
        {
            "level": "resolved",
            "state": "initial",
            "value": "",
            "type": "end_of_text",
        },
    ],
}

# Define a variable which can be later used in a script
SEARCH_NAME = $search_name  # Insert part of the filter name to be searched for

# Filters management
# Get list of filters, with examples of different conditions
list_of_all_filters = list_filters(connection=conn)
list_of_limited_filters = list_filters(connection=conn, limit=10)
list_of_limited_filters_to_dict = list_filters(connection=conn, limit=10, to_dictionary=True)
list_of_filters_by_name = list_filters(connection=conn, name=SEARCH_NAME)

# list of filters with expressions represented as tree
list_of_filters_as_trees = list_filters(connection=conn, show_expression_as=ExpressionFormat.TREE)
# list of filters with expressions represented as tokens
list_of_filters_as_tokens = list_filters(
    connection=conn, show_expression_as=ExpressionFormat.TOKENS
)

# list of filters with qualification represented also in tokens format
list_of_filters_with_qualification_as_tokens = list_filters(
    connection=conn, show_filter_tokens=True
)

# Define variables which can be later used in a script
FILTER_ID = $filter_id  # Insert ID of existing filter here
FILTER_NAME = $filter_name  # Insert name of a new filter to be created

# Get specific filter by id with expressions represented as trees (default)
filter = Filter(connection=conn, id=FILTER_ID)

# Get specified filter by id with expressions represented as tokens
filter_tokens_expression = Filter(
    connection=conn, id=FILTER_ID, show_expression_as=ExpressionFormat.TOKENS
)

# Get specified filter by id with qualification also in tokens format
filter_tokens_qualification = Filter(connection=conn, id=FILTER_ID, show_filter_tokens=True)

# Get filter by name
filter = Filter(connection=conn, name=FILTER_NAME)

# List filter properties
properties = filter.list_properties()

# Define a variable which can be later used in a script
FOLDER_ID = $folder_id  # Insert folder ID here

# create new empty filter
test_filter = Filter.create(
    connection=conn,
    name=FILTER_NAME,
    destination_folder=FOLDER_ID,
)

# create new filter using qualification as tree and
# specify to return qualification also in tokens format
test_filter_tree = Filter.create(
    connection=conn,
    name=(FILTER_NAME + "_tree"),
    destination_folder=FOLDER_ID,
    show_filter_tokens=True,
    qualification=FILTER_QUALIFICATION_AS_TREE,
)

# create new filter using qualification as tokens
test_filter_tokens = Filter.create(
    connection=conn,
    name=(FILTER_NAME + "_tokens"),
    destination_folder=FOLDER_ID,
    qualification=FILTER_QUALIFICATION_AS_TOKENS,
)

# create new filter using qualification as string expression
NEW_FILTER_NAME_STRING = $new_filter_name_string
NEW_FILTER_STRING_QUALIFICATION = "Item Count>1"

new_filter_string = Filter.create(
    connection=conn,
    name=NEW_FILTER_NAME_STRING,
    destination_folder=FOLDER_ID,
    qualification=NEW_FILTER_STRING_QUALIFICATION,
)
print(new_filter_string)

# Any changes to schema object must be followed by schema_reload
# in order to use them in reports, dashboards and so on
schema_manager = SchemaManagement(connection=conn, project_id=conn.project_id)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])

# Define variables which can be later used in a script
FILTER_NEW_NAME = $filter_new_name  # Insert new name of edited Filter here
FILTER_NEW_DESC = $filter_new_desc  # Insert new description of edited filter here

# Alter filter
test_filter.alter(name=FILTER_NEW_NAME, description=FILTER_NEW_DESC)

# Delete newly created filter
test_filter.delete(force=True)



from typing import TYPE_CHECKING

from mstrio.api import library
from mstrio.connection import Connection
from mstrio.object_management import search_operations
from mstrio.object_management.folder import get_folder_id_from_path
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups import User
from mstrio.utils.certified_info import CertifiedInfo
from mstrio.utils.entity import CertifyMixin, CopyMixin, DeleteMixin, Entity, MoveMixin
from mstrio.utils.helper import (
    filter_params_for_func,
    find_object_with_name,
)
from mstrio.utils.library import LibraryMixin
from mstrio.utils.resolvers import (
    get_project_id_from_params_set,
    validate_owner_key_in_filters,
)
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import class_version_handler, method_version_handler

if TYPE_CHECKING:
    from mstrio.server.project import Project


@method_version_handler('11.4.0300')
def _list_implementation(
    factory_class: 'type[Entity]',
    connection: Connection,
    name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    **filters,
):
    from mstrio.object_management.search_enums import SearchPattern

    proj_id = get_project_id_from_params_set(
        connection,
        project,
        project_id,
        project_name,
    )

    validate_owner_key_in_filters(filters)

    objects = search_operations.full_search(
        connection=connection,
        object_types=factory_class._OBJECT_SUBTYPES,
        pattern=SearchPattern.EXACTLY,
        project=proj_id,
        name=name,
        limit=limit,
        **filters,
    )

    if to_dictionary:
        return objects
    else:
        return [
            factory_class.from_dict(source=obj, connection=connection)
            for obj in objects
        ]


def list_agents(
    connection: Connection,
    name: str | None = None,
    to_dictionary: bool = False,
    limit: int | None = None,
    project: 'Project | str | None' = None,
    project_id: str | None = None,
    project_name: str | None = None,
    **filters,
) -> list['Agent'] | list[dict]:
    """Get a list of Agents.

    Args:
        connection (Connection): Strategy One connection object returned
            by 'connection.Connection()'
        name (str, optional): characters that the dashboard name must contain
        to_dictionary (bool, optional): if True, return Agents as a
            list of dicts
        limit (int, optional): limit the number of elements returned.
            If `None` (default), all objects are returned.
        project (Project | str, optional): Project object or ID or name
            specifying the project. May be used instead of `project_id` or
            `project_name`.
        project_id (str, optional): Project ID
        project_name (str, optional): Project name
        **filters: Available filter parameters: ['name', 'id', 'type',
            'subtype', 'date_created', 'date_modified', 'version',
            'owner', 'ext_type', 'view_media', 'certified_info']

    Returns:
        A list of Agent objects or dictionaries.
    """
    return _list_implementation(
        Agent,
        connection=connection,
        name=name,
        to_dictionary=to_dictionary,
        limit=limit,
        project=project,
        project_id=project_id,
        project_name=project_name,
        **filters,
    )


@class_version_handler('11.4.0300')
class Agent(Entity, CertifyMixin, CopyMixin, DeleteMixin, MoveMixin, LibraryMixin):
    """Python representation of a Strategy One Agent object"""

    _OBJECT_TYPE = ObjectTypes.DOCUMENT_DEFINITION

    _OBJECT_SUBTYPES = [
        ObjectSubTypes.DOCUMENT_AGENT,
        ObjectSubTypes.DOCUMENT_AGENT_UNIVERSAL,
    ]

    _API_GETTERS = {
        **Entity._API_GETTERS,
        'status': objects_processors.get_info,
        'recipients': library.get_document,
    }
    _API_PATCH: dict = {
        **Entity._API_PATCH,
        ('status', 'folder_id', 'owner'): (objects_processors.update, 'partial_put'),
    }
    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'certified_info': CertifiedInfo.from_dict,
        'recipients': [User.from_dict],
    }

    def __init__(
        self, connection: Connection, name: str | None = None, id: str | None = None
    ) -> None:
        """Initialize Agent object by passing name or id.

        Args:
            connection (object): Strategy One connection object returned
                by `connection.Connection()`
            name (string, optional): name of Agent
            id (string, optional): ID of Agent
        """
        if id is None:
            if name is None:
                raise ValueError("Please specify either 'name' or 'id'.")

            agent = find_object_with_name(
                connection=connection,
                cls=self.__class__,
                name=name,
                listing_function=list_agents,
            )
            id = agent['id']
        super().__init__(
            connection=connection,
            object_id=id,
            name=name,
        )

    def _init_variables(self, default_value, **kwargs) -> None:
        super()._init_variables(default_value=default_value, **kwargs)
        self._status = kwargs.get('status')

    def alter(
        self,
        name: str | None = None,
        description: str | None = None,
        abbreviation: str | None = None,
        folder_id: str | None = None,
        folder_path: str | None = None,
        hidden: bool | None = None,
        status: str | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
    ) -> None:
        """Alter the Agent.

        Args:
            name (str, optional): New name for the Agent
            description (str, optional): New description for the Agent
            abbreviation (str, optional): New abbreviation for the Agent
            folder_id (str, optional): ID of the folder where the Agent
                should be moved to
            folder_path (str, optional): Path to the folder where the Agent
                should be moved to, optional instead of folder ID
            hidden (bool, optional): Hidden status of the Agent
            status (str, optional): Status of the Agent
                Can be either enabled or disabled
            comments (str, optional): New long description for the Agent
            owner: (str, User, optional): owner of the Driver
        """
        if isinstance(owner, User):
            owner = owner.id
        description = description or self.description
        if folder_path and not folder_id:
            folder_id = get_folder_id_from_path(
                connection=self.connection, path=folder_path
            )
        properties = filter_params_for_func(
            self.alter, locals(), exclude=['self', 'folder_path']
        )
        self._alter_properties(**properties)

    def enable(self) -> None:
        """Enable the Agent."""
        self.alter(status='enabled')

    def disable(self) -> None:
        """Disable the Agent."""
        self.alter(status='disabled')

    @property
    def status(self) -> str:
        """Status of the Agent."""
        if self._status is None:
            self.fetch('status')
        return self._status



from requests import Response

from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error getting privileges.")
def get_privileges(connection, error_msg=None):
    """Get the set of available privileges for the platform.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/iserver/privileges',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting privilege categories.")
def get_privilege_categories(connection, error_msg=None):
    """Get the set of available privilege categories for the platform.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/iserver/privileges/categories',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting LDAP information.")
def get_ldap_info(connection, error_msg=None):
    """Get the existing LDAP configuration and attributes which are stored in
    the Intelligence Server metadata. You may update and save it back later
    based on the existing configuration.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/iserver/ldap/configuration',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error creating LDAP configuration.")
def create_ldap_config(connection, body, error_msg=None):
    """Create a new LDAP configuration and save it to the Intelligence Server.
    There is only one LDAP configuration in Intelligence Server metadata, so
    this operation will replace the old configuration if exists. The request
    body should include all configuration fields.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        endpoint='/api/iserver/ldap/configuration',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error deleting LDAP configuration.")
def delete_ldap_config(connection, error_msg=None):
    """Delete the existing LDAP configuration from Intelligence Server
    metadata.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.delete(
        endpoint='/api/iserver/ldap/configuration',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error updating LDAP configuration.")
def update_ldap_config(connection, body, error_msg=None):
    """Update the existing LDAP configuration and save it to the Intelligence
    Server. There is only one LDAP configuration in Intelligence Server
    metadata, so this operation will replace the old configuration. The request
    body should only contain fields that will be updated.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.patch(
        endpoint='/api/iserver/ldap/configuration',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error getting LDAP attributes.")
def get_ldap_attributes(connection, error_msg=None):
    """Get LDAP attributes list. You may want to import more attributes for
    users. The API will provides you the list of available LDAP attributes,
    from which you may choose some extra attributes to import.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/iserver/ldap/attributes',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting LDAP certificate information.")
def get_ldap_certificate_info(connection, error_msg=None):
    """Get LDAP server certificate info. The communication between the
    Intelligence Server and LDAP server can be encrypted by SSL (optional).
    This API will return basic information about the LDAP server certificate.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/iserver/ldap/certificate',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error updating LDAP certificate.")
def upload_ldap_certificate(connection, body, error_msg=None):
    """Upload LDAP server certificate for encrypted communication. The
    communication between the Intelligence Server and the LDAP server can be
    encrypted by SSL (optional). The certificate uploaded in this API will be
    added to the trust store of the Intelligence Server, and later used for the
    SSL communication between the Intelligence Server and the LDAP server.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        endpoint='/api/iserver/ldap/certificate',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error verifying LDAP binding.")
def verify_ldap_binding(connection, error_msg=None):
    """Verify the LDAP username and password binding.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/iserver/ldap/authuserbinds',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error getting LDAP batch import status.")
def get_ldap_batch_import_status(connection, error_msg=None):
    """Get LDAP batch import status. You can get the progress of LDAP batch
    import, including its status (e.g. failed, stopped, undergoing, and
    canceled). If there is LDAP batch import undergoing, you can also get the
    number of users and groups which have already been imported.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/iserver/ldap/import',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error stopping LDAP batch import.")
def stop_ldap_batch_import(connection, error_msg=None):
    """Stop the LDAP batch import if it exists. This operation will be ignored
    if there is no undergoing LDAP batch import.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint='/api/iserver/ldap/import',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error doing LDAP batch import.")
def do_ldap_batch_import(connection, error_msg=None):
    """Do LDAP batch import now. This API starts the LDAP batch import, which
    will import both groups and users from the LDAP server. No request body is
    needed.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        endpoint='/api/iserver/ldap/import',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error updating authentication configurations.")
def update_authentication_configs(connection, body, error_msg=None):
    """Update the authentication configuration settings for the current REST
    server. In the body parameter of the request, you specify the default
    authentication mode and the authentication modes that will be available to
    connect to an Intelligence Server.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling


    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint='/api/admin/restServerSettings/auth',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
        json=body,
    )


@ErrorHandler(err_msg="Error updating collaboration server configurations.")
def update_collaboration_server_configs(connection, body, error_msg=None):
    """Update the Collaboration Server configuration settings. In the body
    parameter of the request, you specify the base URL for the Collaboration
    Server, whether the collaboration service is enabled, and whether TLS
    (Transport Layer Security) secure communication is enabled. Enabling TLS
    requires configuring trust store settings on the REST Server.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint='/api/admin/restServerSettings/collaboration',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
        json=body,
    )


@ErrorHandler(err_msg="Error updating web configurations.")
def update_web_configs(connection, body, error_msg=None):
    """Update the Strategy One Web configuration settings. In the body
    parameter of the request, you specify the base URL for the Strategy One
    Web Server.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint='/api/admin/restServerSettings/microStrategyWeb',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
        json=body,
    )


@ErrorHandler(err_msg="Error testing collaboration server connection.")
def test_collaboration_connection(connection, body, error_msg=None):
    """Test the connection between the REST Server and the Collaboration
    Server. In the body parameter of the request, you specify the base URL for
    the Collaboration Server and whether TLS (Transport Layer Security) secure
    communication is enabled. Enabling TLS requires configuring trust store
    settings on the REST Server.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        endpoint='/api/admin/restServerSettings/collaboration/connectionTest',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
        json=body,
    )


@ErrorHandler(err_msg="Error updating intelligence server configurations.")
def update_iserver_configs(connection, body, error_msg=None):
    """Update the Intelligence Server configuration settings. In the body
    parameter of the request, you specify the default port and hostname, the
    initial and max connection pool size, the request timeout, the number of
    reports/documents and searches that can be kept in memory, whether a trust
    relationship has been set with the Intelligence Server.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint='/api/admin/restServerSettings/iServer',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
        json=body,
    )


@ErrorHandler(err_msg="Error getting rest configurations.")
def get_rest_configs(connection, error_msg=None):
    """Get configuration settings for the current REST Server, including
    settings for authentication, the Collaboration Server, Google Analytics,
    and the Intelligence Server.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/admin/restServerSettings',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
    )


@ErrorHandler(
    err_msg="Error checking intelligence server - web server trust relationship."
)
def check_iserver_web_trust(connection, error_msg=None):
    """Check to see if there is a trust relationship between the Web Server and
    the Intelligence Server.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/admin/restServerSettings/iServer/trustRelationship',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
    )


@ErrorHandler(
    err_msg="Error setting intelligence server - web server trust relationship."
)
def set_iserver_web_trust(connection, body, error_msg=None):
    """Set up a trust relationship between the Web Server and the Intelligence
    Server. You obtain the authorization token needed to execute the request
    using POST /auth/login; you pass the authorization token in the request
    header. In the body parameter of the request, you specify the URL or other
    unique identifier for the Web Server.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        endpoint='/api/admin/restServerSettings/iServer/trustRelationship',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
        json=body,
    )


@ErrorHandler(
    err_msg="Error deleting intelligence server - web server trust relationship."
)
def delete_iserver_web_trust(connection, error_msg=None):
    """Delete a trust relationship between the Web Server and the Intelligence
    Server. You obtain the authorization token needed to execute the request
    using POST /auth/login; you pass the authorization token in the request
    header.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.delete(
        endpoint='/api/admin/restServerSettings/iServer/trustRelationship',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
    )


@ErrorHandler(err_msg="Error testing intelligence server connection.")
def test_iserver_connection(connection, error_msg=None):
    """Test the connection between the Intelligence Server and the REST Server.
    Currently, only socket connectivity is tested. In the body parameter of the
    request, you specify the port and hostname of the Intelligence Server.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.post(
        endpoint='/api/admin/restServerSettings/iServer/connectionTest',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
    )


@ErrorHandler(err_msg="Error getting security settings.")
def get_security_settings(connection, error_msg=None):
    """Get relevant information about security settings, such as whether the
    secret key used for signing the identity token was set in the configuration
    file. If it was not set properly, an exception will be triggered.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/admin/restServerSettings/security',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
    )


@ErrorHandler(err_msg="Error updating security settings.")
def update_security_settings(connection, body, error_msg=None):
    """Update the security settings. In the body of the request you can specify
    the secret key that will be used in identity token generation, whether to
    enable Cross Origin Resource Sharing (CORS), and whether to disable URL
    validation when specifying domains for CORS configuration.

    Args:
        connection: Strategy One REST API connection object
        body: JSON-formatted definition of the dataset. Generated by
            `utils.formjson()`.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint='/api/admin/restServerSettings/security',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
        json=body,
    )


@ErrorHandler(err_msg="Error fetching I-Server settings.")
def get_iserver_settings_config(connection, error_msg=None):
    """For each setting, we provide different basic information to help client
    to render the settings.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/v2/iserver/settings/config',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error fetching I-Server settings.")
def get_iserver_settings(connection, error_msg=None):
    """This resource will retire resource 'GET iserver/settings' in the future
    version. Current version just cover Governing rule settings.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/v2/iserver/settings',
        headers={'X-MSTR-ProjectID': None},
    )


@ErrorHandler(err_msg="Error updating I-Server settings.")
def create_iserver_settings(
    connection, body, error_msg=None, whitelist=('ERR001', 400)
):
    """Create some IServer governing settings.

    Example:
        body = {
            "maxUsedVirtualByte": {
                "value": 100
            },
            "workSetSwapPath": {
                "value": ".\\TmpPool"
            },
            "ignoreGovernorOnChild": {
                "value": true
            }
        }
    """
    return connection.put(
        endpoint='/api/v2/iserver/settings',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error updating I-Server settings.")
def update_iserver_settings(
    connection, body, error_msg=None, whitelist=('ERR001', 400)
):
    """Update some IServer governing settings.

    Example:
        body = {
            "maxUsedVirtualByte": {
                "value": 100
            },
            "workSetSwapPath": {
                "value": ".\\TmpPool"
            },
            "ignoreGovernorOnChild": {
                "value": true
            }
        }
    """
    return connection.patch(
        endpoint='/api/v2/iserver/settings',
        headers={'X-MSTR-ProjectID': None},
        json=body,
    )


@ErrorHandler(err_msg="Error getting cluster membership information.")
def get_cluster_membership(connection, error_msg=None):
    """Get IServer cluster membership information.

    Args:
        connection: Strategy One REST API connection object
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/admin/iServer/clusterMembership',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
    )


@ErrorHandler(err_msg="Error getting I-Server node {node} settings.")
def get_iserver_node_settings(connection, node, error_msg=None):
    """Get Intelligence Server configuration settings for a given server node
    within a cluster.

    Args:
        connection: Strategy One REST API connection object
        node: Intelligence Server host name
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/admin/restServerSettings/iServer/{node}',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
    )


@ErrorHandler(err_msg="Error updating Intelligence Server configuration settings.")
def update_iserver_configuration_settings(connection, body, error_msg=None):
    """Update Intelligence Server configuration settings.

    Args:
        connection: Strategy One REST API connection object
        body: object with settings to set
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint='/api/admin/restServerSettings/iServer',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
        json=body,
    )


@ErrorHandler(err_msg="Error updating I-Server node {node} settings.")
def update_iserver_node_settings(connection, body, node, error_msg=None):
    """Update Intelligence Server configuration settings for a given server
    node within a cluster.

    Args:
        connection: Strategy One REST API connection object
        node: Intelligence Server host name
        error_msg (string, optional): Custom Error Message for Error Handling
        body:{"loadBalanceFactor": int,
              "initialPoolSize": int,
              "maxPoolSize": int}

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint=f'/api/admin/restServerSettings/iServer/{node}',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
        json=body,
    )


@ErrorHandler(err_msg="Error deleting I-Server node {node} settings.")
def delete_iserver_node_settings(connection, node, error_msg=None):
    """Removes Intelligence Server configuration settings for a given server
    node within a cluster. This can be useful to clear out settings so that a
    default value is applied.

    Args:
        connection: Strategy One REST API connection object
        node: Intelligence Server host name
        error_msg (string, optional): Custom Error Message for Error Handling
    """
    return connection.delete(
        endpoint=f'/api/admin/restServerSettings/iServer/{node}',
        headers={
            'X-MSTR-ProjectID': None,
            'Authorization': connection._get_authorization(),
        },
    )


@ErrorHandler(err_msg="Error getting Library storage configuration.")
def storage_service_get_configs(
    connection: Connection,
    fields: str | None = None,
    error_msg: str | None = None,
):
    """Get the storage type, location, and cloud storage configurations.

    Args:
        connection (Connection): Strategy One REST API connection object
        fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. 200 on success.
    """
    return connection.get(
        endpoint='/api/mstrServices/library/storage', params={'fields': fields}
    )


@ErrorHandler(err_msg="Error validating Library storage configuration.")
def storage_service_validate_configs(
    connection: Connection, body: dict, error_msg: str | None = None
):
    """Validate the storage type, location, and cloud storage configurations.
    Following privilege: DssXmlPrivilegesAdministerEnvironment is required to
    invoke this endpoint.

    Args:
        connection (Connection): Strategy One REST API connection object
        body (dict): {
            "sharedFileStore": {
                "type": str
                    ("unset", "unknown", "file_system", "S3", "Azure", "GCS"),
                "alias": str (optional),
                "location": str (optional),
                "s3Region": str (optional),
                "awsAccessId": str (optional),
                "awsSecretKey": str (optional),
                "azureStorageAccountName": str (optional),
                "azureSecretKey": str (optional),
                "gcsServiceAccountKey": str (optional)
            }
        }
            "sharedFileStore"/"type" supports:
            "file_system", "S3", "Azure", "GCS"
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. 204 on success.
    """
    return connection.post(
        endpoint='/api/mstrServices/library/storage/validation',
        json=body,
    )


@ErrorHandler(err_msg="Error updating Library storage configuration.")
def storage_service_update_configs(
    connection: Connection, body: dict, error_msg: str | None = None
):
    """Update the storage type, location, and cloud storage configurations.
    Following privilege: DssXmlPrivilegesAdministerEnvironment is required to
    invoke this endpoint.

    Args:
        connection (Connection): Strategy One REST API connection object
        body (dict): {
            "sharedFileStore": {
                "type": str
                    ("unset", "unknown", "file_system", "S3", "Azure", "GCS"),
                "alias": str (optional),
                "location": str (optional),
                "s3Region": str (optional),
                "awsAccessId": str (optional),
                "awsSecretKey": str (optional),
                "azureStorageAccountName": str (optional),
                "azureSecretKey": str (optional),
                "gcsServiceAccountKey": str (optional)
            }
        }
            "sharedFileStore"/"type" supports:
            "file_system", "S3", "Azure", "GCS"
        error_msg (string, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object. 204 on success.
    """
    return connection.patch(
        endpoint='/api/mstrServices/library/storage',
        json=body,
    )


@ErrorHandler(err_msg="Error getting fences list.")
def list_fences(
    connection: Connection, fields: str | None = None, error_msg: str | None = None
) -> Response:
    """Get the list of fences.

    Args:
        connection (Connection): Strategy One REST API connection object
        fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(endpoint='/api/v2/iserver/fences', params={'fields': fields})


@ErrorHandler(err_msg="Error creating a new fence.")
def create_fence(
    connection: Connection,
    body: dict,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Create a new fence.

    Args:
        connection (Connection): Strategy One REST API connection object
        body (dict): Fence creation data
        fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.

    """
    return connection.post(
        endpoint='/api/v2/iserver/fences',
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg="Error getting fence with ID: {id}")
def get_fence(
    connection: Connection,
    id: str,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Get a fence by ID.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (str): Fence ID
        fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint=f'/api/v2/iserver/fences/{id}',
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error deleting fence with ID: {id}")
def delete_fence(
    connection: Connection, id: str, error_msg: str | None = None
) -> Response:
    """Delete a fence by ID.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (str): Fence ID
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.delete(endpoint=f'/api/v2/iserver/fences/{id}')


@ErrorHandler(err_msg="Error updating fence with ID: {id}")
def update_fence(
    connection: Connection,
    id: str,
    body: dict,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Update a fence by ID.

    Args:
        connection (Connection): Strategy One REST API connection object
        id (str): Fence ID
        body (dict): Fence update data
        fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.patch(
        endpoint=f'/api/v2/iserver/fences/{id}',
        params={'fields': fields},
        json=body,
    )


@ErrorHandler(err_msg="Error getting cluster startup membership information.")
def get_cluster_startup_membership(
    connection: Connection, fields: str | None = None, error_msg: str | None = None
) -> Response:
    """Get the cluster startup membership information.

    Args:
        connection (Connection): Strategy One REST API connection object
        fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.get(
        endpoint='/api/iserver/clusterStartupMembership',
        params={'fields': fields},
    )


@ErrorHandler(err_msg="Error updating cluster startup membership information.")
def update_cluster_startup_membership(
    connection: Connection,
    body: dict,
    fields: str | None = None,
    error_msg: str | None = None,
) -> Response:
    """Update the cluster startup membership information.

    Args:
        connection (Connection): Strategy One REST API connection object
        body (dict): Cluster startup membership update data
        fields (str, optional): A comma-separated list of fields to include
                in the response. By default, all fields are returned.
        error_msg (str, optional): Custom Error Message for Error Handling

    Returns:
        Complete HTTP response object.
    """
    return connection.put(
        endpoint='/api/iserver/clusterStartupMembership',
        json=body,
        params={'fields': fields},
    )



"""
Manage email subscriptions after cube refresh.
Select a cube, refresh, find its subscriptions and run them.

1. Connect to the environment using data from the workstation
2. Get the intelligent cube object based on provided id
3. Publish retrieved cube and wait for its refresh to complete successfully
4. Get the list of dashboards dependent on the cube
5. Get the list of all subscriptions
6. For each found dashboard, find all subscriptions using it and store them
   in a list
7. Trigger each stored subscription
"""

import time
from mstrio.connection import get_connection
from mstrio.project_objects.datasets.cube import CubeStates
from mstrio.project_objects.datasets import OlapCube
from mstrio.distribution_services.subscription.subscription_manager import (
    list_subscriptions,
)


conn = get_connection(workstationData, 'MicroStrategy Tutorial')


# Select an intelligent cube and refresh it
example_cube = OlapCube(
    connection=conn, id="8CCD8D9D4051A4C533C719A6590DEED8"
)  # Intelligent Cube - Drilling
example_cube.publish()


# Wait for cube refresh to complete successfully.
example_cube.refresh_status()
while "Processing" in CubeStates.show_status(example_cube.status):
    time.sleep(1)
    example_cube.refresh_status()


# Search for subscriptions dependent on that cube
# (find dashboards dependent on cube, and then subscriptions
# dependent on those dashboards found)
dependent_dashboards = example_cube.list_dependents()
all_subscriptions = list_subscriptions(conn)
subscriptions_to_send = []
for d in dependent_dashboards:
    # For each found Dashboard, find all Subscriptions using it.
    # This step requires filtering a list of Subscriptions with Python
    # list operations, as Document/Dashboard query responses do not backlink
    # to Subscriptions.
    dashboard_id = d['id']

    subscriptions_to_send += [
        sub
        for sub in all_subscriptions
        if any(content.id == dashboard_id for content in sub.contents)
    ]

# Trigger (run) the subscriptions
for s in subscriptions_to_send:
    s.execute()



from mstrio.connection import Connection
from mstrio.utils.api_helpers import changeset_manager, unpack_information
from mstrio.utils.error_handlers import ErrorHandler


@unpack_information
@ErrorHandler(err_msg="Error creating an attribute")
def create_attribute(
    connection: Connection,
    body: dict,
    show_expression_as: list[str] | None = None,
    show_potential_tables: str | None = None,
    show_fields: str | None = None,
    fields: str | None = None,
):
    """Create a new attribute in the changeset,
    based on the definition provided in request body.

    Args:
        connection: Strategy One REST API connection object
        body: Attribute creation data
        show_expession_as: Specifies the format in which the expressions are
           returned in response
           Available values: 'tokens', 'tree'
           If omitted, the expression is returned in 'text' format
           If 'tree', the expression is returned in 'text' and 'tree' formats.
           If 'tokens', the expression is returned in 'text' and 'tokens'
           formats.
        show_potential_tables: Specifies whether to return the potential tables
            that the expressions can be applied to.
            Available values: 'true', 'false'
            If 'true', the 'potentialTables' field returns for each attribute
                expression, in the form of a list of tables.
            If 'false' or omitted, the 'potentialTables' field is omitted.
        show_fields: Specifies what additional information to return.
            Only 'acl' is supported.
        fields: A whitelist of top-level fields separated by commas.
            Allow the client to selectively retrieve fields in the response.

    Return:
        HTTP response object. Expected status: 201
    """
    with changeset_manager(connection) as changeset_id:
        return connection.post(
            endpoint='/api/model/attributes',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            params={
                'showExpressionAs': show_expression_as,
                'showPotentialTables': show_potential_tables,
                'showFields': show_fields,
                'fields': fields,
            },
            json=body,
        )


@unpack_information
@ErrorHandler(err_msg="Error getting attribute with ID: {id}")
def get_attribute(
    connection: Connection,
    id: str,
    changeset_id: str | None = None,
    show_expression_as: list[str] | None = None,
    show_potential_tables: str | None = None,
    show_fields: str | None = None,
    fields: str | None = None,
):
    """Get definition of a single attribute by id

    Args:
        connection: Strategy One REST API connection object
        id: ID of an attribute
        changeset_id: ID of a changeset
        show_expession_as: Specifies the format in which the expressions
           are returned in response
           Available values: 'tokens', 'tree'
           If omitted, the expression is returned in 'text' format
           If 'tree', the expression is returned in 'text' and 'tree' formats.
           If 'tokens', the expression is returned in 'text' and 'tokens'
           formats.
        show_potential_tables: Specifies whether to return the potential tables
            that the expressions can be applied to.
            Available values: 'true', 'false'
            If 'true', the 'potentialTables' field returns for each
                attribute expression, in the form of a list of tables.
            If 'false' or omitted, the 'potentialTables' field is omitted.
        show_fields: Specifies what additional information to return.
            Only 'acl' is supported.
        fields: A whitelist of top-level fields separated by commas.
            Allow the client to selectively retrieve fields in the response.

    Return:
        HTTP response object. Expected status: 200
    """
    connection._validate_project_selected()
    return connection.get(
        endpoint=f'/api/model/attributes/{id}',
        headers={'X-MSTR-MS-Changeset': changeset_id},
        params={
            'showExpressionAs': show_expression_as,
            'showPotentialTables': show_potential_tables,
            'showFields': show_fields,
            'fields': fields,
        },
    )


@unpack_information
@ErrorHandler(err_msg="Error updating attribute with ID: {id}")
def update_attribute(
    connection: Connection,
    id: str,
    body: dict,
    show_expression_as: list[str] | None = None,
    show_potential_tables: str | None = None,
    show_fields: str | None = None,
    fields: str | None = None,
    remove_invalid_fields: str | None = None,
):
    """Update a specific attribute in the changeset
    This endpoint replaces the attribute's top-level fields
    with the new definition provided in the request body.

    Args:
        connection: Strategy One REST API connection object
        id: ID of an attribute
        body: Attribute update data
        show_expression_as: Specifies the format in which the expressions
           are returned in response.
           Available values: 'tokens', 'tree'
           If omitted, the expression is returned in 'text' format
           If 'tree', the expression is returned in 'text' and 'tree' formats.
           If 'tokens', the expression is returned in 'text' and 'tokens'
           formats.
        show_potential_tables: Specifies whether to return the potential tables
            that the expressions can be applied to.
            Available values: 'true', 'false'
            If 'true', the 'potentialTables' field returns for each
                attribute expression, in the form of a list of tables.
            If 'false' or omitted, the 'potentialTables' field is omitted.
        show_fields: Specifies what additional information to return.
            Only 'acl' is supported.
        fields: A whitelist of top-level fields separated by commas.
            Allow the client to selectively retrieve fields in the response.
        remove_invalid_fields: Specifies whether to check and remove the invalid
            fields caused by changes related to key form within the request.

    Return:
        HTTP response object. Expected status: 200
    """
    with changeset_manager(connection) as changeset_id:
        return connection.patch(
            endpoint=f'/api/model/attributes/{id}',
            headers={'X-MSTR-MS-Changeset': changeset_id},
            params={
                'showExpressionAs': show_expression_as,
                'showPotentialTables': show_potential_tables,
                'showFields': show_fields,
                'fields': fields,
                'removeInvalidFields': remove_invalid_fields,
            },
            json=body,
        )


@ErrorHandler(err_msg="Error getting attribute with ID: {id}")
def get_attribute_elements(
    connection: Connection,
    id: str,
    fields: str | None = None,
):
    """Get definition of a single attribute by ID.

    Args:
        connection: Strategy One REST API connection object
        id: ID of an attribute
        fields: A whitelist of top-level fields separated by commas.
            Allow the client to selectively retrieve fields in the response.

    Return:
        HTTP response object. Expected status: 200
    """
    connection._validate_project_selected()
    return connection.get(
        endpoint=f'/api/attributes/{id}/elements',
        params={
            'fields': fields,
        },
    )



import logging
from concurrent.futures import Future, as_completed
from functools import cached_property

from requests import HTTPError, ReadTimeout
from tqdm import tqdm

from mstrio.api import datasources
from mstrio.api import tables as tables_api
from mstrio.connection import Connection
from mstrio.datasources.datasource_instance import (
    DatasourceInstance,
    list_connected_datasource_instances,
)
from mstrio.modeling.schema import ObjectSubType, SchemaObjectReference
from mstrio.modeling.schema.helpers import TableColumn, TableColumnMergeOption
from mstrio.modeling.schema.table.logical_table import LogicalTable, list_logical_tables
from mstrio.utils.helper import Dictable, fetch_objects, get_response_json
from mstrio.utils.sessions import FuturesSessionWithRenewal
from mstrio.utils.version_helper import method_version_handler

logger = logging.getLogger(__name__)


@method_version_handler('11.3.0100')
def list_datasource_warehouse_tables(
    connection: Connection,
    datasource_id: str,
    namespace_id: str,
    name: str = None,
    refresh: bool = False,
    to_dictionary: bool = False,
    limit: int | None = None,
) -> list[type["WarehouseTable"]] | list[dict]:
    """Lists available warehouse tables in a specified datasource within a
       specified namespace.

    Args:
        connection (Connection): Object representation of MSTR Connection.
        datasource_id (str): ID of a datasource.
        namespace_id (str): ID of a namespace within a given datasource_id.
        name (str): string by which to filter the table name
        refresh (bool, optional): Refresh warehouse table, default: False.
        to_dictionary (bool, optional): If True, the function will return list
            of dictionaries. Otherwise, list of WarehouseTable objects.
            Defaults to False.
        limit (int, optional): limit the number of elements returned. If `None`
            (default), all objects are returned.

    Returns:
        Union[list["WarehouseTable"], list[dict]]: A list of WarehouseTable
            or dictionary objects representing warehouse tables in a specified
            namespace in a specified datasource.

    Examples:
        >>> datasource_id = "DAD6CAD6457DAF29E34463961688EA60"
        >>> namespace_id = "eyJucyI6InB1YmxpYyJ9"
        >>> available_wh_tables = list_datasource_warehouse_tables(
            connection, datasource_id, namespace_id)
    """

    tables = fetch_objects(
        connection=connection,
        api=tables_api.get_available_warehouse_tables,
        limit=limit,
        filters={'name': name} if name else None,
        dict_unpack_value="tables",
        datasource_id=datasource_id,
        namespace_id=namespace_id,
        refresh=refresh,
    )
    [
        table.update(
            {"datasource": {"id": datasource_id}, "namespace_id": namespace_id}
        )
        for table in tables
    ]
    if to_dictionary:
        return tables
    return WarehouseTable.bulk_from_dict(source_list=tables, connection=connection)


def list_warehouse_tables(
    connection: Connection,
    to_dictionary: bool = False,
    name: str | None = None,
    datasource_id: str | None = None,
) -> list["WarehouseTable"] | list[dict]:
    """Fetches all available warehouse table. This operation is done
       asynchronously and is heavy: a lot of requests are performed to fetch
       all connected and available datasources, namespaces and finally
       warehouse tables for each combination. Additionally, the result could be
       filtered. Available filters are table name and datasource id. These
       filters can be combined.

    Args:
        connection (Connection): Object representation of MSTR Connection.
        to_dictionary (bool, optional): If True, returns a list of dictionaries.
            If False, returns a list of WarehouseTable objects.
        name (str, optional): A name of a warehouse table.
        datasource_id (str, optional): ID of a datasource.

    Returns:
        list[WarehouseTable] | list[dict]: A list of dictionaries
            or WarehouseTable objects.

    """
    return WarehouseTable._list_warehouse_tables(
        connection, to_dictionary, name, datasource_id
    )


class WarehouseTable(Dictable):
    """An object representation of a warehouse table. Depending on whether a
    warehouse table is included in a project, it may or may not have a
    persistent id and associated PhysicalTable object. When you add a warehouse
    table to a project, a MSTR server will create a PhysicalTable object that
    represents the added warehouse table and will assign it a persistent ID.
    You can then manipulate the table further using functionalities in
    mstrio.modeling.schema.table.physical_table.

    Attributes:
        connection (Connection): An object representation of MSTR Connection.
        datasource (DatasourceInstance): An object representation of datasource
            instance in which the table resides.
        name (str): A name of the table.
        namespace (str): A name of a namespace in which the table resides.
        namespace_id (str): An ID of a namespace in which the table resides.
        physical_table_id(str): An ID of an associated PhysicalTable object.
            This attribute is set automatically after `add_to_project()` method
            was called on the table.
        dependent_logical_tables (Union[list[dict], list[LogicalTable]]): A list
            of logical tables that are mapped to the warehouse table.
        columns (Union[list[dict], list[TableColumn]]): A list of table columns.

    """

    _FROM_DICT_MAP: dict = {"datasource": DatasourceInstance.from_dict}

    def __init__(
        self,
        id: str,
        connection: Connection,
        datasource: DatasourceInstance,
        namespace_id: str,
        name: str,
        namespace: str,
    ):
        self.__id = (
            id  # Non-persistent. It will change if the table name or namespace changes.
        )
        self.connection = connection
        self.datasource = datasource
        self.name = name
        self.namespace = namespace
        self.namespace_id = namespace_id
        # When table is added to a project, it will get a persistent ID
        self.physical_table_id = None
        self._dependent_logical_tables = None
        self._columns = None

    def add_to_project(
        self,
        logical_table_name: str,
        prefix: str | None = None,
        col_merge_option: TableColumnMergeOption | str | None = None,
    ) -> type["LogicalTable"]:
        """Adds a table to a project. This function corresponds to 'ADD WHTABLE'
           statement from MSTR Command Manager.

        Args:
            logical_table_name (str): Name of the logical table which will be
                mapped to this warehouse table.
            prefix (Optional[str], optional): Table prefix. Defaults to None.
            col_merge_option (Optional[Union[TableColumnMergeOption, str]]:
                Defines a column merge option. Defaults to None.

        Returns:
            LogicalTable: A LogicalTable object representing a logical table
                mapped to this warehouse table.
        """
        lt = LogicalTable.create(
            connection=self.connection,
            table_name=logical_table_name,
            primary_data_source=SchemaObjectReference(
                object_id=self.datasource.id, sub_type=ObjectSubType.LOGICAL_TABLE
            ),
            column_merge_option=col_merge_option,
            physical_table_prefix=prefix,
            physical_table_name=self.name,
            physical_table_namespace=self.namespace,
        )
        self.physical_table_id = lt.physical_table.id
        self.list_dependent_logical_tables(refresh=True)
        return lt

    def list_dependent_logical_tables(
        self, to_dictionary: bool = False, refresh: bool = False
    ) -> list["LogicalTable"] | list[dict]:
        """Get all dependent logical tables.

        Args:
            to_dictionary (bool, optional): If True, returns a list of
                dictionaries. Defaults to False.
            refresh(bool, optional): If True, refreshes the result stored in
                `self.dependent_logical_tables`.

        Returns:
            Union[list[LogicalTable], list[dict]]: A list of LogicalTable
                objects or dictionaries representing logical tables.
        """
        if self._dependent_logical_tables and not refresh:
            return self._dependent_logical_tables
        logical_tables = list_logical_tables(connection=self.connection)
        if self.physical_table_id:
            dependent_logical_tables = [
                table
                for table in logical_tables
                if table.physical_table.id == self.physical_table_id
            ]
        else:
            dependent_logical_tables = [
                table
                for table in logical_tables
                if table.physical_table.table_name == self.name
            ]

        self._dependent_logical_tables = dependent_logical_tables

        if to_dictionary:
            return [table.to_dict() for table in dependent_logical_tables]

        return dependent_logical_tables

    def delete_from_project(self, force: bool = False) -> list[bool] | None:
        """Remove the Warehouse Table from a project by removing all of its
        dependent logical tables. It will only delete dependent logical
        tables if they themselves have no dependent objects. This function
        corresponds to "REMOVE WHTABLE" statement from MSTR Command Manager.
        """
        dependent_logical_tables = self.list_dependent_logical_tables(refresh=True)
        if dependent_logical_tables:
            logger.warning("Following logical tables will be deleted: ")
            [logger.info(f"{str(table)}") for table in dependent_logical_tables]
            confirmed_delete = (
                force or input("Would you like to continue? Y/n").lower() == "y"
            )
            if confirmed_delete:
                statuses = [
                    table.delete(force=True) for table in dependent_logical_tables
                ]
                self.physical_table_id = (
                    None if all(statuses) else self.physical_table_id
                )
                return statuses
        else:
            logger.error("This table is not included in a project.")

    def list_columns(
        self, to_dictionary: bool = False, refresh: bool = False
    ) -> list[type[TableColumn]] | list[dict]:
        """Get columns for a specific database table.

        Args:
            to_dictionary (bool, optional): If True, returns a list of
                dictionaries. Defaults to False.
            refresh (bool, optional): If True, refetches and overwrites the
                cached version of results stored in
                WarehouseTable.columns.

        Returns:
            Union[list[TableColumn], list[dict]]: A list of TableColumn objects
                or dictionaries representing table columns.
        """
        if self._columns and not refresh:
            return self._columns
        datasource_id = (
            self.datasource.id
            if isinstance(self.datasource, DatasourceInstance)
            else self.datasource.get("id")
        )
        columns = fetch_objects(
            connection=self.connection,
            api=datasources.get_table_columns,
            limit=None,
            filters={},
            dict_unpack_value="columns",
            datasource_id=datasource_id,
            namespace_id=self.namespace_id,
            table_id=self.__id,
        )

        if to_dictionary:
            return columns

        column_objects = TableColumn.bulk_from_dict(
            source_list=columns, connection=self.connection
        )
        self._columns = column_objects
        return column_objects

    @classmethod
    def _list_warehouse_tables(
        cls,
        connection: Connection,
        to_dictionary: bool = False,
        name: str | None = None,
        datasource_id: str | None = None,
    ) -> list["WarehouseTable"] | list[dict]:
        """Fetches all available warehouse table in a project mapped to the
           Connection object. This operation is done asynchronously and is
           heavy: a lot of requests are performed to fetch all connected and
           available datasources, namespaces and finally warehouse tables for
           each combination. Additionally, the result could be filtered.
           Available filters are table name and datasource id. These
           filters can be combined.

        Args:
            connection (Connection): Object representation of MSTR Connection.
            to_dictionary (bool, optional): If True, returns a list of
                dictionaries. If False, returns a list of WarehouseTable
                objects.
            name (str, optional): A name of a warehouse table.
            datasource_id (str, optional): ID of a datasource.

        Returns:
            list[WarehouseTable] | list[dict]: A list of dictionaries
                or WarehouseTable objects.
        """
        if name or datasource_id:
            tables = cls._filter(
                connection=connection, name=name, datasource_id=datasource_id
            )
            if to_dictionary:
                return [table.to_dict() for table in tables]
            return tables

        return cls._list_available_warehouse_tables(
            connection=connection, to_dictionary=to_dictionary
        )

    @classmethod
    def _filter(
        cls,
        connection: Connection,
        name: str | None = None,
        datasource_id: str | None = None,
    ) -> list["WarehouseTable"]:
        """Fetches (if not yet fetched) and filters all available warehouse
           tables. Available filters are table name and datasource id. These
           filters can be combined.

           Args:
            connection (Connection): Object representation of MSTR Connection.
            name (str, optional): A name of a warehouse table.
            datasource_id (str, optional): ID of a datasource.

        Returns:
            list[WarehouseTable]: A list of filtered warehouse tables.
        """

        available_tables = cls._list_available_warehouse_tables(connection)

        if name and datasource_id:
            datasource_tables = [
                table
                for table in available_tables
                if table.datasource.id == datasource_id
            ]
            available_tables = [
                table for table in datasource_tables if table.name == name.lower()
            ]

        elif name:
            available_tables = [
                table for table in available_tables if table.name == name.lower()
            ]
        elif datasource_id:
            available_tables = [
                table
                for table in available_tables
                if table.datasource.id == datasource_id
            ]
        return available_tables

    @classmethod
    def _list_available_warehouse_tables(
        cls, connection: Connection, to_dictionary: bool = False
    ) -> list["WarehouseTable"] | list[dict]:
        """Fetches all available warehouse table in a project mapped to the
           Connection object. This operation is done asynchronously and is
           heavy: a lot of requests are performed to fetch all connected and
           available datasources, namespaces and finally warehouse tables for
           each combination.

           connection (Connection): Object representation of MSTR Connection.
           to_dictionary (bool): If True, returns a list of dictionaries.

        Returns:
            Union[list[dict], list["WarehouseTable"]]: A list of dictionaries
                or WarehouseTable objects.
        """

        connected_datasource_instances: list[dict] = (
            list_connected_datasource_instances(connection, to_dictionary=True)
        )
        endpoints: dict[str, str] = cls._get_namespaces_endpoints(
            connected_datasource_instances
        )

        with FuturesSessionWithRenewal(connection=connection) as session:
            namespaces: dict[str, list[dict]] = cls._get_namespaces(endpoints, session)

            warehouse_tables_futures = cls._get_warehouse_tables_futures(
                session, namespaces
            )

            warehouse_tables: list[dict] = cls._get_warehouse_tables(
                warehouse_tables_futures, session
            )

            available_tables = cls.bulk_from_dict(
                source_list=warehouse_tables, connection=connection
            )
        if to_dictionary:
            return warehouse_tables

        return available_tables

    @classmethod
    def _get_warehouse_tables(
        cls, warehouse_tables_futures: list[Future], session: FuturesSessionWithRenewal
    ) -> list[dict]:
        """Retrieves warehouse tables from a list of provided futures.

        Args:
            warehouse_tables_futures (list[Future]): A list of futures that
            resolve to warehouse tables in a given datasource in a given
            namespace.

        Returns:
            list[dict]: A list of warehouse tables retrieved from all the
                futures.
        """

        warehouse_tables: list[dict] = []
        with tqdm(
            total=len(warehouse_tables_futures), desc="Retrieving warehouse tables..."
        ) as pbar:
            for future in as_completed(warehouse_tables_futures):
                namespace_tables = (
                    cls._get_future_with_request_exceptions_handlers_and_pbar(
                        cls._get_tables_from_future, future, pbar, session=session
                    )
                )
                if namespace_tables:
                    warehouse_tables += namespace_tables

        return warehouse_tables

    @classmethod
    def _get_namespaces(
        cls, endpoints: dict, session: FuturesSessionWithRenewal
    ) -> dict[str, list[dict]]:
        """Retrieves namespaces for every url using provided session object.

        Args:
            endpoints (dict): A dictionary with datasource ID as a key and
                list of namespace endpoints as a value.
            session (FuturesSession): A FuturesSession object used to fetch
                urls asynchronously.

        Returns:
            dict[str, list[dict]]: A dictionary which keys are datasource ids
                and which values are lists of namespaces.
        """
        namespaces: dict[str, list[dict]] = {}
        namespaces_futures = cls._get_namespaces_futures(session, endpoints)
        with tqdm(
            total=len(namespaces_futures),
            desc="Retrieving namespaces from available datasources...",
        ) as pbar:
            for future in as_completed(namespaces_futures):
                namespaces[future.datasource_id] = (
                    cls._get_future_with_request_exceptions_handlers_and_pbar(
                        cls._get_namespaces_from_future, future, pbar
                    )
                )
        return namespaces

    @classmethod
    def _get_future_with_request_exceptions_handlers_and_pbar(
        cls, func, future, pbar, **func_kwargs
    ):
        try:
            return func(future=future, **func_kwargs)
        except (ReadTimeout, HTTPError, ConnectionError) as e:
            logger.error(e)
        finally:
            pbar.update()

    @staticmethod
    def _get_namespaces_endpoints(
        connected_datasource_instances: list[dict],
    ) -> dict[str, str]:
        """Creates urls to api/datasources/{datasource_id}/catalog/namespaces
           that are later used to fetch all namespaces from a specified
           datasource

        Args:
            connected_datasource_instances (list[dict]): list of dictionaries
                representing connected datasource instances.

        Returns:
            dict[str, str]: a dictionary with datasource id as a key,
                and matching namespaces url as a value
        """
        return {
            connected_datasource_instance.get("id"): (
                f"/api/datasources/"
                f"{connected_datasource_instance.get('id')}"
                f"/catalog/namespaces"
            )
            for connected_datasource_instance in connected_datasource_instances
        }

    @staticmethod
    def _get_namespaces_futures(
        session: FuturesSessionWithRenewal, endpoints: dict[str, str]
    ) -> list[Future]:
        """Creates Future objects using specified FuturesSession object for each
           url.

        Args:
            session (FuturesSession): A FuturesSession object used to fetch
                the url.
            endpoints (dict[str, str]): a dictionary with datasource id
                as a key, and matching namespaces url as a value.

        Returns:
            list[Future]: A list of Future objects that later resolve to
                namespaces within a specified datasource.
        """
        futures = []
        for datasource_id, endpoint in endpoints.items():
            future = session.get(
                endpoint=endpoint,
                headers={"X-MSTR-ProjectID": session.connection.project_id},
            )
            future.datasource_id = datasource_id
            futures.append(future)
        return futures

    @staticmethod
    def _get_warehouse_tables_futures(
        session: FuturesSessionWithRenewal, namespaces: dict[str, list[dict]]
    ) -> list[Future]:
        """Creates Future objects using specified FuturesSession object for each
           namespace for each datasource.

        Args:
            session (FuturesSession): A FuturesSession object used to fetch
                the tables.
            namespaces (dict[str, list[dict]]): a dictionary with datasource id
                as a key and list of available namespaces within this
                datasource as a value.

        Returns:
            list[Future]: A list of Future objects that later resolve to
                warehouse tables for each namespace for each datasource.
        """
        futures = []
        for datasource_id, namespaces_list in namespaces.items():
            if namespaces_list:
                for namespace in namespaces_list:
                    namespace_id = namespace.get("id")
                    future = tables_api.get_available_warehouse_tables_async(
                        session, datasource_id=datasource_id, namespace_id=namespace_id
                    )
                    future.datasource_id = datasource_id
                    future.namespace_id = namespace_id
                    future.namespace = namespace.get("name")
                    futures.append(future)
        return futures

    @staticmethod
    def _get_namespaces_from_future(future: Future) -> list[dict]:
        """Retrieves a list of namespaces from a given future.

        Args:
            future (Future): A future object that resolves to a dictionary that
                contains a list of namespaces.

        Returns:
            list[dict]: A list of namespaces retrieved from the future.
        """
        resp = future.result()
        data = get_response_json(resp, throw_error=False)
        if isinstance(data, dict):
            return data.get("namespaces", [])
        return []

    @staticmethod
    def _get_tables_from_future(
        future: Future, session: FuturesSessionWithRenewal
    ) -> list[dict]:
        """Retrieves a list of tables from a given future.

        Args:
            future (Future): A future object that resolves to a dictionary that
                contains a list of tables.
            session (FuturesSessionWithRenewal): A FuturesSessionWithRenewal
                object used to fetch the tables.

        Returns:
            list[dict]: A list of tables retrieved from the future.
        """
        warehouse_tables = []
        resp = future.result()
        tables = get_response_json(resp).get("tables")
        if tables:
            for table in tables:
                table["connection"] = session.connection
                table["datasource"] = {"id": future.datasource_id}
                table["namespace_id"] = future.namespace_id
                warehouse_tables.append(table)
        return warehouse_tables

    @property
    def dependent_logical_tables(self):
        if not self._dependent_logical_tables:
            self._dependent_logical_tables = self.list_dependent_logical_tables()
        return self._dependent_logical_tables

    @cached_property
    def columns(self):
        self._columns = self.list_columns()
        return self._columns



"""This is the demo script meant to show how to control the refresh status of
OLAP cube.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from pprint import pprint
from time import sleep

from mstrio.connection import get_connection
from mstrio.helpers import IServerError
from mstrio.project_objects import OlapCube, list_olap_cubes
from mstrio.server import Job, JobStatus


def inform_when_job_done(job: Job, interval: int = 10):
    """Helper function to check the status of the job every `interval` seconds
    until it is completed. It will print the error message if the job is
    completed with error. This works only when called in short time max 20
    minutes after triggering the refresh."""
    stable_states = [JobStatus.COMPLETED, JobStatus.ERROR, JobStatus.STOPPED]
    try:
        while job.status not in stable_states:
            sleep(interval)
            job.fetch()
        print(f"Job is finished with error: {job.error_message}")
    except IServerError:
        print("Job is finished with success.")

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn = get_connection(workstationData, project_name=PROJECT_NAME)


# List available OLAP cubes
cubes = list_olap_cubes(connection=conn)
# Print the list of available OLAP cubes
pprint(cubes)

# To get existing OLAP cube you need to know its ID
SELECTED_CUBE_ID = $selected_cube_id  # Insert ID of cube on which you want to perform actions
selected_cube = OlapCube(connection=conn, id=SELECTED_CUBE_ID)
# Print the last update time of the selected OLAP cube
before_last_update_time = selected_cube.last_update_time
print(before_last_update_time)
# Print the status of the selected OLAP cube
before_status = selected_cube.status
print(selected_cube.show_status())
# Refresh the selected OLAP cube
job = selected_cube.refresh()

# Now request for refresh was sent to the I-Server. It may take some time to
# complete the refresh. You can check the status of the job to see if it is
# completed or see if last update time has changed.

# Job can be taken from the job object or initialized with the job ID
JOB_ID = job.id
my_job = Job(connection=conn, id=JOB_ID)

# Here we will check the status of the job every 10 seconds until it is
# completed. You can adjust the time interval to your needs.

inform_when_job_done(job=my_job, interval=10)

# Another way to check job status is by calling refresh_status method on the job
# If the job can not be found on server when refreshing, the status will be
# updated to JobStatus.COMPLETED
my_job.refresh_status()

# Now you can check the last update time of the selected OLAP cube to see if
# the refresh was successful.
selected_cube.fetch()
after_last_update_time = selected_cube.last_update_time
print(after_last_update_time)
# Print the status of the selected OLAP cube
after_status = selected_cube.status
print(selected_cube.show_status())

# Other way to check if refresh is done is comparing last update time before
# and after refresh. If they are different, the refresh was successful.
if before_last_update_time != after_last_update_time:
    print("Refresh was successful.")



"""This is the demo script to show how administrator can manage shortcuts.

This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to
ease its usage.
"""

from mstrio.connection import get_connection
from mstrio.modeling.metric import Metric
from mstrio.object_management.shortcut import list_shortcuts


# Create connection based on workstation data
PROJECT_NAME = $project_name  # Project to connect to
conn = get_connection(workstationData, project_name=PROJECT_NAME)

# Create a shortcut for an existing object, e.g. a metric
METRIC_ID = $metric_id
metric_object = Metric(conn, id=METRIC_ID)

TARGET_FOLDER_ID = $target_folder_id
shortcut = metric_object.create_shortcut(target_folder_id=TARGET_FOLDER_ID)

# List shortcuts
shortcuts = list_shortcuts(conn)
for sc in shortcuts:
    print(sc)

# list properties for a shortcut
properties = shortcut.list_properties()
for prop_name, prop_value in properties.items():
    print(f"{prop_name}: {prop_value}")

# alter properties of a shortcut
SHORTCUT_NEW_NAME = $shortcut_new_name
shortcut.alter(name=SHORTCUT_NEW_NAME)

# delete a shortcut
shortcut.delete(force=True)



from mstrio.connection import Connection
from mstrio.utils.error_handlers import ErrorHandler


@ErrorHandler(err_msg="Error reading lock status of the schema.")
def read_lock_status(connection: 'Connection', project_id: str | None = None):
    """Read lock status of the schema."""
    project_id = project_id if project_id is not None else connection.project_id
    return connection.get(
        endpoint='/api/model/schema/lock',
        headers={'X-MSTR-ProjectID': project_id},
    )


@ErrorHandler(err_msg="Error placing the lock of type `{lock_type}` on the schema.")
def lock_schema(
    connection: 'Connection',
    lock_type: str,
    project_id: str | None = None,
    throw_error: bool = True,
):
    """Places a lock on the schema."""
    project_id = project_id if project_id is not None else connection.project_id
    return connection.post(
        endpoint='/api/model/schema/lock',
        headers={'X-MSTR-ProjectID': project_id},
        json={'lockType': lock_type},
    )


@ErrorHandler(err_msg="Error unlocking the schema.")
def unlock_schema(
    connection: 'Connection',
    lock_type: str | None = None,
    project_id: str | None = None,
    throw_error: bool = True,
):
    """Unlocks the schema."""
    project_id = project_id if project_id is not None else connection.project_id
    return connection.delete(
        endpoint='/api/model/schema/lock',
        headers={'X-MSTR-ProjectID': project_id},
        params={'lockType': lock_type},
    )


@ErrorHandler(err_msg="Error reloading the schema.")
def reload_schema(
    connection: 'Connection',
    project_id: str | None = None,
    update_types: list[str] | None = None,
    prefer_async: bool = False,
):
    """Reloads (updates) the schema."""
    project_id = project_id if project_id is not None else connection.project_id
    update_types = update_types if update_types else []
    prefer_async = 'respond-async' if prefer_async else None
    return connection.post(
        endpoint='/api/model/schema/reload',
        headers={
            'X-MSTR-ProjectID': project_id,
            'Prefer': prefer_async,
        },
        json={'updateTypes': update_types},
    )


@ErrorHandler(err_msg="Error reading status of the task with ID: {task_id}.")
def read_task_status(
    connection: 'Connection', task_id: str, project_id: str | None = None
):
    """Read the status of the task."""
    project_id = project_id if project_id is not None else connection.project_id
    return connection.get(
        endpoint=f'/api/model/tasks/{task_id}',
        headers={'X-MSTR-ProjectID': project_id},
    )



from logging import getLogger

from mstrio import config
from mstrio.connection import Connection
from mstrio.types import ObjectSubTypes, ObjectTypes
from mstrio.users_and_groups.user import User
from mstrio.utils import helper
from mstrio.utils.entity import CopyMixin, DeleteMixin, Entity
from mstrio.utils.enums import DaysOfWeek
from mstrio.utils.response_processors import calendars as calendars_processors
from mstrio.utils.response_processors import objects as objects_processors
from mstrio.utils.version_helper import class_version_handler, method_version_handler

logger = getLogger(__name__)


@method_version_handler(version='11.3.0700')
def list_calendars(
    connection: Connection,
    to_dictionary: bool = False,
    limit: int | None = None,
    offset: int | None = None,
    only_system_calendars: bool | None = None,
    only_custom_calendars: bool | None = None,
):
    """List all available calendar objects.

    Args:
        connection (Connection): Strategy One connection object returned
            by `connection.Connection()`
        to_dictionary: If True, returns a list of dicts,
            otherwise returns a list of Calendar objects
        limit (int, optional): Limit the number of calendars returned
        offset (int, optional): Starting point within the collection of
            returned results. Used to control paging behavior.
        only_system_calendars (bool, optional): If True, returns only system
            calendars. Cannot be used with `only_custom_calendars`
        only_custom_calendars (bool, optional): If True, returns only custom
            calendars. Cannot be used with `only_system_calendars`

    Returns:
    A list of calendars as Calendar objects or dicts."""

    return Calendar._list_all(
        connection=connection,
        to_dictionary=to_dictionary,
        limit=limit,
        offset=offset,
        only_system_calendars=only_system_calendars,
        only_custom_calendars=only_custom_calendars,
    )


@class_version_handler(version='11.3.0700')
class Calendar(Entity, CopyMixin, DeleteMixin):

    _OBJECT_TYPE = ObjectTypes.CALENDAR
    _API_GETTERS = {
        (
            'name',
            'description',
            'type',
            'subtype',
            'ext_type',
            'date_created',
            'date_modified',
            'version',
            'owner',
            'ancestors',
            'acg',
            'acl',
            'comments',
            'hidden',
        ): objects_processors.get_info,
        (
            'id',
            'base_calendar',
            'calendar_begin_static',
            'calendar_begin_offset',
            'calendar_end_static',
            'calendar_end_offset',
            'table_prefix',
            'week_start_day',
        ): calendars_processors.get_calendar,
    }
    _API_PATCH = {
        (
            'name',
            'description',
            'table_prefix',
            'base_calendar',
            'calendar_begin',
            'calendar_end',
            'week_start_day',
        ): (calendars_processors.update_calendar, 'partial_put'),
        (
            'comments',
            'owner',
        ): (objects_processors.update, 'partial_put'),
    }
    _PATCH_PATH_TYPES = {
        **Entity._PATCH_PATH_TYPES,
        'base_calendar': dict,
    }

    @staticmethod
    def _parse_base_calendar(
        source, connection, to_snake_case: bool = True
    ) -> 'Calendar':
        """Parses base calendar from the API response."""

        return Calendar.from_dict(source, connection, to_snake_case)

    @classmethod
    def _validate_year_definition_deltas(
        cls,
        calendar_begin_static: int | None,
        calendar_begin_offset: int | None,
        calendar_end_static: int | None,
        calendar_end_offset: int | None,
    ) -> None:
        if calendar_begin_static is not None and calendar_begin_offset is not None:
            raise ValueError(
                "You must provide either 'calendar_begin_static' "
                "or 'calendar_begin_offset', not both."
            )
        if calendar_end_static is not None and calendar_end_offset is not None:
            raise ValueError(
                "You must provide either 'calendar_end_static' or "
                "'calendar_end_offset', not both."
            )

    @classmethod
    def _validate_year_definitions(
        cls,
        calendar_begin_static: int | None,
        calendar_begin_offset: int | None,
        calendar_end_static: int | None,
        calendar_end_offset: int | None,
    ) -> None:
        cls._validate_year_definition_deltas(
            calendar_begin_static=calendar_begin_static,
            calendar_begin_offset=calendar_begin_offset,
            calendar_end_static=calendar_end_static,
            calendar_end_offset=calendar_end_offset,
        )
        if calendar_begin_static is None and calendar_begin_offset is None:
            raise ValueError(
                "You must provide either 'calendar_begin_static' "
                "or 'calendar_begin_offset'."
            )
        if calendar_end_static is None and calendar_end_offset is None:
            raise ValueError(
                "You must provide either 'calendar_end_static' "
                "or 'calendar_end_offset'."
            )

    _FROM_DICT_MAP = {
        **Entity._FROM_DICT_MAP,
        'subtype': ObjectSubTypes,
        'owner': User.from_dict,
        'week_start_day': DaysOfWeek,
    }

    def __init__(
        self,
        connection: Connection,
        id: str | None = None,
        name: str | None = None,
    ) -> None:
        if not id:
            if name:
                # not using helper functions because they're suited to Response
                # objects and not extracted/wrangled dicts returned by
                # response_processors functions
                cals_with_name = [
                    c
                    for c in Calendar._list_all(
                        connection=connection, to_dictionary=True
                    )
                    if c['name'] == name
                ]
                if not cals_with_name:
                    raise ValueError(f"Calendar with name '{name}' not found.")
                if (num_of_cals := len(cals_with_name)) > 1:
                    raise ValueError(
                        f"Found {num_of_cals} calendars with name '{name}'. "
                        "Please provide 'id' argument."
                    )
                id = cals_with_name[0]['id']
            else:
                raise ValueError("Please provide either 'id' or 'name' argument.")
        super().__init__(connection=connection, object_id=id, name=name)

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.base_calendar = kwargs.get('base_calendar')
        self.table_prefix = kwargs.get('table_prefix')
        if calendar_begin_static := kwargs.get('calendar_begin_static'):
            self.calendar_begin_static = int(calendar_begin_static)
        if calendar_begin_offset := kwargs.get('calendar_begin_offset'):
            self.calendar_begin_offset = int(calendar_begin_offset)
        if calendar_end_static := kwargs.get('calendar_end_static'):
            self.calendar_end_static = int(calendar_end_static)
        if calendar_end_offset := kwargs.get('calendar_end_offset'):
            self.calendar_end_offset = int(calendar_end_offset)
        if week_start_day := kwargs.get('week_start_day'):
            self.week_start_day = DaysOfWeek(week_start_day.lower())

    @classmethod
    def create(
        cls,
        connection: Connection,
        name: str,
        base_calendar: 'str | Calendar',
        week_start_day: DaysOfWeek,
        table_prefix: str,
        description: str | None = None,
        calendar_begin_static: int | None = None,
        calendar_begin_offset: int | None = None,
        calendar_end_static: int | None = None,
        calendar_end_offset: int | None = None,
    ) -> 'Calendar':
        """Create a new calendar with the specified properties.

        Args:
            connection (Connection): Strategy One connection object returned by
                `connection.Connection()`
            name (str): Name of the calendar object
            base_calendar (str or Calendar, optional): Reference (ID or Calendar
                object) to the base calendar
            week_start_day (DaysOfWeek): First day of the week
            table_prefix (str): Prefix used to create tables in the warehouse.
                Defaults to an empty string
            description (str, optional): Description of the calendar object
            calendar_begin_static (int, optional): Beginning year for the
                calendar. Must be provided if and only if
                `calendar_begin_offset` is not
            calendar_begin_offset (int, optional): Beginning year for the
                calendar as an offset from the current year. Must be provided
                if and only if `calendar_begin_static` is not
            calendar_end_static (int, optional): Ending year for the calendar.
                Must be provided if and only if `calendar_end_offset` is not
            calendar_end_offset (int, optional): Ending year for the calendar
                as an offset from the current year. Must be provided if and
                only if `calendar_end_static` is not
        """

        cls._validate_year_definitions(
            calendar_begin_static=calendar_begin_static,
            calendar_begin_offset=calendar_begin_offset,
            calendar_end_static=calendar_end_static,
            calendar_end_offset=calendar_end_offset,
        )

        base_calendar_id = (
            base_calendar.id if isinstance(base_calendar, cls) else base_calendar
        )
        body = {
            "information": {"name": name, "description": description},
            "tablePrefix": table_prefix,
            "baseCalendar": {
                "objectId": base_calendar_id,
            },
            "calendarBegin": {
                "staticYear": calendar_begin_static,
                "dynamicYearOffset": calendar_begin_offset,
            },
            "calendarEnd": {
                "staticYear": calendar_end_static,
                "dynamicYearOffset": calendar_end_offset,
            },
            "weekStartDay": week_start_day.value,
        }

        body = helper.delete_none_values(source=body, recursion=True)
        new_cal = cls.from_dict(
            source=calendars_processors.create_calendar(
                connection=connection, body=body
            ),
            connection=connection,
        )
        if config.verbose:
            logger.info(
                f"Successfully created Calendar named: '{name}' with ID:"
                f" '{new_cal.id}'"
            )
        return new_cal

    def alter(
        self,
        name: str | None = None,
        week_start_day: DaysOfWeek | None = None,
        description: str | None = None,
        comments: str | None = None,
        owner: str | User | None = None,
        base_calendar: 'str | Calendar | None' = None,
        table_prefix: str | None = None,
        calendar_begin_static: int | None = None,
        calendar_begin_offset: int | None = None,
        calendar_end_static: int | None = None,
        calendar_end_offset: int | None = None,
    ) -> None:
        """Alter the calendar's properties.
        Args:
            name (str, optional): Name of the calendar object
            week_start_day (DaysOfWeek, optional): First day of the week
            description (str, optional): Description of the calendar object
            comments (str, optional): long description of the calendar object
            owner: (str or User, optional): owner of the calendar object
            base_calendar (str or Calendar, optional): Reference (ID or Calendar
                object) to the base calendar
            table_prefix (str, optional): Prefix used to create tables
                in the warehouse
            calendar_begin_static (int, optional): Beginning year for the
                calendar. Must be provided if and only if
                `calendar_begin_offset` is not
            calendar_begin_offset (int, optional): Beginning year for the
                calendar as an offset from the current year. Must be provided
                if and only if `calendar_begin_static` is not
            calendar_end_static (int, optional): Ending year for the calendar.
                Must be provided if and only if `calendar_end_offset` is not
            calendar_end_offset (int, optional): Ending year for the calendar
                as an offset from the current year. Must be provided if and
                only if `calendar_end_static` is not
        """
        if isinstance(owner, User):
            owner = owner.id
        self._validate_year_definition_deltas(
            calendar_begin_static=calendar_begin_static,
            calendar_begin_offset=calendar_begin_offset,
            calendar_end_static=calendar_end_static,
            calendar_end_offset=calendar_end_offset,
        )

        if base_calendar is None:
            base_calendar_id = self.base_calendar
        elif isinstance(base_calendar, Calendar):
            base_calendar_id = base_calendar.id
        else:
            base_calendar_id = base_calendar

        name = name or self.name
        table_prefix = table_prefix or self.table_prefix
        if calendar_begin_static is None and calendar_begin_offset is None:
            calendar_begin_static = getattr(self, 'calendar_begin_static', None)
            calendar_begin_offset = getattr(self, 'calendar_begin_offset', None)
        if calendar_end_static is None and calendar_end_offset is None:
            calendar_end_static = getattr(self, 'calendar_end_static', None)
            calendar_end_offset = getattr(self, 'calendar_end_offset', None)
        week_start_day = week_start_day or self.week_start_day

        properties = {
            "name": name,
            "description": description,
            "comments": comments,
            "owner": owner,
            "table_prefix": table_prefix,
            "base_calendar": {"object_id": base_calendar_id},
            "calendar_begin": {
                "static_year": calendar_begin_static,
                "dynamic_year_offset": calendar_begin_offset,
            },
            "calendar_end": {
                "static_year": calendar_end_static,
                "dynamic_year_offset": calendar_end_offset,
            },
            "week_start_day": week_start_day.value,
        }

        properties = helper.delete_none_values(properties, recursion=True)
        self._alter_properties(**properties)

    def get_calendar_begin(self, base_year: int | None) -> int:
        if self.calendar_begin_offset is None:
            return self.calendar_begin_static
        else:
            if base_year is None:
                raise ValueError(
                    'Base year is required to calculate the year with offset.'
                )
            return base_year + self.calendar_begin_offset

    def get_calendar_end(self, base_year: int | None) -> int:
        if self.calendar_end_offset is None:
            return self.calendar_end_static
        else:
            if base_year is None:
                raise ValueError(
                    'Base year is required to calculate the year with offset.'
                )
            return base_year + self.calendar_end_offset

    @classmethod
    def _list_all(
        cls,
        connection: Connection,
        to_dictionary: bool = False,
        limit: int | None = None,
        offset: int | None = None,
        only_system_calendars: bool | None = None,
        only_custom_calendars: bool | None = None,
    ) -> list['Calendar'] | list[dict]:
        if only_system_calendars and only_custom_calendars:
            raise ValueError(
                "You can use only one "
                "of 'only_system_calendars' and 'only_custom_calendars' arguments."
            )

        subtype = None
        if only_system_calendars:
            subtype = "calendar_system"
        elif only_custom_calendars:
            subtype = "calendar_custom"

        objects = calendars_processors.list_calendars(
            connection=connection,
            limit=limit,
            offset=offset,
            subtype=subtype,
        )
        if to_dictionary:
            return objects
        else:
            return [cls.from_dict(source=obj, connection=connection) for obj in objects]



from typing import TYPE_CHECKING

from mstrio.api import datasources
from mstrio.datasources.datasource_login import DatasourceLogin
from mstrio.datasources.helpers import CharEncoding, DriverType, ExecutionMode
from mstrio.utils.entity import EntityBase
from mstrio.utils.helper import get_args_from_func, get_default_args_from_func
from mstrio.utils.version_helper import class_version_handler

if TYPE_CHECKING:
    from mstrio.connection import Connection


@class_version_handler('11.3.0000')
class EmbeddedConnection(EntityBase):
    """Datasource connection configuration object that represents an embedded
    connection template for the datasource.

    Attributes:
        connection: A Strategy One connection object.
        id: Unique datasource connection ID.
        name: Unique datasource connection name.
        datasource_id: ID of the corresponding datasource instance.
        description: Datasource connection description.
        execution_mode: ExecutionMode Enum specifying how SQL statements will be
            executed on a physical database connection made.
        max_cancel_attempt_time: Number of seconds before being timed out when
            attempting to make a connection.
        max_query_exe_time: Number of seconds during which a query must be
            executed before being timed out.
        max_connection_attempt_time: Number of seconds connection attempts to be
            made before timing out.
        connection_lifetime: Number of seconds of the connection lifetime.
        connection_idle_timeout: Number of seconds before being timed out when
            the connection is idle.
        char_encoding_windows: CharEncoding Enum specifying the encoding
            across connection for Windows drivers.
        char_encoding_unix: CharEncoding Enum specifying the encoding across
            connection for Unix drivers.
        table_prefix: String prefixed to the names of all temporary tables
            created by the Query Engine for this connection during report
            execution with this string.
        connection_string: Database connection string.
        parameterized_queries: Specifies whether parameterized queries are
            enabled.
        extended_fetch: Specifies whether or not to use the extended fetch ODBC
            call to retrieve data from the database connection.
        datasource_login: `DatasourceLogin` object or ID
        database_type: Database type
        database_version: Database version
        driver_type: DriverType Enum specifying Driver used for database
            connection.
        oauth_parameter: Used for authentication with oAuth.
        type: Object type
        subtype: Object subtype
        date_created: Creation time, DateTime object
        date_modified: Last modification time, DateTime object
        owner: User object that is the owner
        acg: Object access rights. It a bit vector where bits are defined at
            [EnumDSSXMLAccessRightFlags].
        acl: Object access control list
        iam: List of projects to which the fence is applied.
        resource: The url of configured Web API for OAuth authentication usage.
        scope: List of delegated permissions that the app is requesting.
        enable_sso: Specifies whether to use Single Sign-On.
    """

    _FROM_DICT_MAP = {
        **EntityBase._FROM_DICT_MAP,
        'execution_mode': ExecutionMode,
        'datasource_login': DatasourceLogin.from_dict,
        'driver_type': DriverType,
        'char_encoding_unix': CharEncoding,
        'char_encoding_windows': CharEncoding,
    }
    _API_GETTERS = {
        (
            'id',
            'name',
            'description',
            'execution_mode',
            'max_cancel_attempt_time',
            'max_query_exe_time',
            'max_connection_attempt_time',
            'connection_lifetime',
            'connection_idle_timeout',
            'char_encoding_windows',
            'char_encoding_unix',
            'table_prefix',
            'connection_string',
            'parameterized_queries',
            'extended_fetch',
            'driver_type',
            'datasource_login',
            'database_type',
            'database_version',
            'oauth_parameter',
            'acg',
            'iam',
            'resource',
            'scope',
            'enable_sso',
        ): datasources.get_embedded_connection,
    }
    _API_PATCH: dict = {
        (
            "description",
            "execution_mode",
            "max_cancel_attempt_time",
            "max_query_exe_time",
            "max_connection_attempt_time",
            "connection_lifetime",
            "connection_idle_timeout",
            "char_encoding_windows",
            "char_encoding_unix",
            "table_prefix",
            "connection_string",
            "parameterized_queries",
            "extended_fetch",
            "driver_type",
            "oauth_parameter",
            'resource',
            'scope',
            'enable_sso',
        ): (
            datasources.update_embedded_connection,
            'patch',
        ),
    }
    _PATCH_PATH_TYPES = {
        "name": str,
        "description": str,
        "execution_mode": str,
        "max_cancel_attempt_time": int,
        "max_query_exe_time": int,
        "max_connection_attempt_time": int,
        "connection_lifetime": int,
        "connection_idle_timeout": int,
        "char_encoding_windows": str,
        "char_encoding_unix": str,
        "table_prefix": str,
        "parameterized_queries": bool,
        "extended_fetch": bool,
        "driver_type": str,
        "oauth_parameter": str,
        "iam": dict,
        "resource": str,
        "scope": str,
        "enable_sso": bool,
    }

    def __init__(
        self,
        connection: "Connection",
        datasource_id: str,
    ) -> None:
        """Initialize EmbeddedConnection object and synchronize with server.

        Args:
            connection: Strategy One connection object returned by
                `connection.Connection()`.
            datasource_id: ID of the corresponding Datasource Connection.
                The embedded connection object has an ID, but it cannot be used
                in a meaningful way in REST API. The corresponding datasource
                instance ID is used instead.
        """

        super().__init__(connection=connection, object_id=datasource_id)
        self.datasource_id = datasource_id

    def _init_variables(self, **kwargs) -> None:
        super()._init_variables(**kwargs)
        self.name = kwargs.get("name")
        self.description = kwargs.get("description")
        self.execution_mode = (
            ExecutionMode(kwargs["execution_mode"])
            if kwargs.get("execution_mode")
            else None
        )
        self.datasource_id = kwargs.get("datasource_id")
        self.max_cancel_attempt_time = kwargs.get("max_cancel_attempt_time")
        self.max_query_exe_time = kwargs.get("max_query_exe_time")
        self.max_connection_attempt_time = kwargs.get("max_connection_attempt_time")
        self.connection_lifetime = kwargs.get("connection_lifetime")
        self.connection_idle_timeout = kwargs.get("connection_idle_timeout")
        self.char_encoding_windows = (
            CharEncoding(kwargs["char_encoding_windows"])
            if kwargs.get("char_encoding_windows")
            else None
        )
        self.char_encoding_unix = (
            CharEncoding(kwargs["char_encoding_unix"])
            if kwargs.get("char_encoding_unix")
            else None
        )
        self.table_prefix = kwargs.get("table_prefix")
        self.connection_string = kwargs.get("connection_string")
        self.parameterized_queries = kwargs.get("parameterized_queries")
        self.extended_fetch = kwargs.get("extended_fetch")
        self.driver_type = (
            DriverType(kwargs["driver_type"]) if kwargs.get("driver_type") else None
        )
        self.datasource_login = (
            DatasourceLogin.from_dict(datasource_login, self.connection)
            if (datasource_login := kwargs.get("datasource_login"))
            else None
        )
        self.database_type = kwargs.get("database_type")
        self.database_version = kwargs.get("database_version")
        self._oauth_parameter = kwargs.get("oauth_parameter")
        self.iam = kwargs.get("iam")
        self.resource = kwargs.get("resource")
        self.scope = kwargs.get("scope")
        self.enable_sso = kwargs.get("enable_sso")

    def alter(
        self,
        description: str | None = None,
        execution_mode: str | ExecutionMode | None = None,
        max_cancel_attempt_time: int | None = None,
        max_query_exe_time: int | None = None,
        max_connection_attempt_time: int | None = None,
        connection_lifetime: int | None = None,
        connection_idle_timeout: int | None = None,
        char_encoding_windows: str | CharEncoding | None = None,
        char_encoding_unix: str | CharEncoding | None = None,
        table_prefix: str | None = None,
        connection_string: str | None = None,
        parameterized_queries: bool | None = None,
        extended_fetch: bool | None = None,
        driver_type: str | DriverType | None = None,
        oauth_parameter: str | None = None,
        resource: str | None = None,
        scope: str | None = None,
        enable_sso: bool | None = None,
        comments: str | None = None,
    ) -> None:
        """Alter the datasource connection properties.

        Args:
            description: Datasource connection description.
            execution_mode: ExecutionMode Enum specifying how SQL statements
                will be executed on a physical database connection made.
            max_cancel_attempt_time: Number of seconds before being timed out
                when attempting to make a connection.
            max_query_exe_time: Number of seconds during which a query must be
                executed before being timed out.
            max_connection_attempt_time: Number of seconds connection attempts
                to be made before timing out.
            connection_lifetime: Number of seconds of the connection lifetime.
            connection_idle_timeout: Number of seconds before being timed out
                when the connection is idle.
            char_encoding_windows: CharEncoding Enum specifying the encoding
                across connection for Windows drivers.
            char_encoding_unix: CharEncoding Enum specifying the encoding across
                connection for Unix drivers.
            table_prefix: String prefixed to the names of all temporary tables
                created by the Query Engine for this connection during report
                execution with this string.
            connection_string: Database connection string.
            parameterized_queries: Specifies whether parameterized queries are
                enabled.
            extended_fetch: Specifies whether or not to use the extended fetch
                ODBC call to retrieve data from the database connection.
            driver_type: DriverType Enum specifying Driver used for database
                connection.
            datasource_login: `DatasourceLogin` object or ID
            driver_type: ENUM Drivers used for database connection.
            database_type: Database type
            database_version: Database version
            resource: The url of configured Web API for OAuth authentication
                usage.
            scope: List of delegated permissions that the app is requesting.
            enable_sso: Specifies whether to use Single Sign-On.
        """
        func = self.alter
        args = get_args_from_func(func)
        defaults = get_default_args_from_func(func)
        default_dict = dict(zip(args[-len(defaults) :], defaults)) if defaults else {}
        local = locals()
        properties = {'datasource_id': self.datasource_id}
        for property_key in default_dict.keys():
            if local[property_key] is not None and property_key != 'datasource_id':
                properties[property_key] = local[property_key]
        self._alter_properties(**properties)

    @property
    def oauth_parameter(self):
        return self._oauth_parameter



from mstrio.utils import helper


class SettingValueFactory:
    def get_setting(self, config: dict) -> "SettingValue":
        setting_value_types = {
            'number': NumberSetting,
            'string': StringSetting,
            'enum': EnumSetting,
            'boolean': BoolSetting,
            'time': TimeSetting,
            'email': EmailSetting,
            'object': ObjectSetting,
            'mstr_object': MstrObjectSetting,
            '_modified': bool,
            None: DeprecatedSetting,
        }
        setting_type = config.get('type')
        return setting_value_types[setting_type](config)


class SettingValue:
    """Base Settings Value class.

    It represents single setting configuration.
    """

    def __init__(self, config: dict):
        self.value = None
        self.name = config.get('name')
        self.description = config.get('description')
        self.reboot_rule = config.get('reboot_rule')
        self.multi_select = config.get('multi_select')
        self.options = config.get('options', [])
        self.relationship = config.get('relationship')
        self._modified = False

    def __repr__(self):
        return str(self._get_value())

    def __setattr__(self, name, value):
        """Setattr that allows setting valid name values defined in options."""
        # if setting the value of setting and options are defined
        # and value is str (option name)
        value_is_option_name = isinstance(value, str) and value
        setting_val_with_options = name == 'value' and getattr(self, 'options', None)
        if setting_val_with_options and value_is_option_name:
            option_found = helper.filter_list_of_dicts(self.options, name=value)
            if option_found:
                value = option_found[0]['value']
        if name == 'value':
            old_value = getattr(self, 'value', None)
            if old_value is not None:
                if old_value != value:
                    self._modified = True
                else:
                    self._modified = False

        super().__setattr__(name, value)

    def _validate_value(self, value, exception=True):
        options = helper.extract_all_dict_values(self.options)
        return helper.validate_param_value(
            self.name, value, self.type, special_values=options, exception=exception
        )

    def _get_value(self):
        return self.value

    @property
    def info(self):
        return self.__dict__


class EnumSetting(SettingValue):
    """Representation of an Enum setting type."""

    # initially empty options that require additional action
    # to be available (e.g. creating a TimeZone object)
    _DYNAMICALLY_ADDED_OPTIONS = {'defaultTimezone': str, 'defaultCalendar': str}

    def __init__(self, config: dict):
        super().__init__(config)
        config_name = config.get('name')
        if config_name in self._DYNAMICALLY_ADDED_OPTIONS:  # noqa: SIM401
            self.type = self._DYNAMICALLY_ADDED_OPTIONS[config_name]
        else:
            self.type = list if self.multi_select else type(self.options[0]['value'])

    def _validate_value(self, value, exception=True):
        options = helper.extract_all_dict_values(self.options)
        if self.type is list:
            options.append('')
        return helper.validate_param_value(
            self.name, value, self.type, special_values=options, exception=exception
        )

    def _get_value(self):
        option_name = [
            option['name']
            for option in helper.filter_list_of_dicts(self.options, value=self.value)
        ]
        if len(option_name) == 1:
            return option_name[0]
        else:
            return option_name


class BoolSetting(SettingValue):
    """Representation of a Boolean setting type."""

    def __init__(self, config: dict):
        super().__init__(config)
        self.type = list if self.multi_select else bool


class NumberSetting(SettingValue):
    """Representation of a Number setting type."""

    def __init__(self, config: dict):
        super().__init__(config)
        self.type = list if self.multi_select else int
        self.max_value = config.get('max_value')
        self.min_value = config.get('min_value')
        self.unit = config.get('unit')

    def _validate_value(self, value, exception=True):
        options = helper.extract_all_dict_values(self.options)
        return helper.validate_param_value(
            self.name,
            value,
            self.type,
            self.max_value,
            self.min_value,
            options,
            exception=exception,
        )


class StringSetting(SettingValue):
    """Representation of a String setting type."""

    def __init__(self, config: dict):
        super().__init__(config)
        self.type = list if self.multi_select else str

    def _validate_value(self, value, exception=True):
        return helper.validate_param_value(
            self.name, value, self.type, exception=exception
        )


class TimeSetting(SettingValue):
    """Representation of a Time setting type."""

    def __init__(self, config: dict):
        super().__init__(config)
        self.type = list if self.multi_select else str

    def _validate_value(self, value, exception=True):
        regex = r"^[1-2][0-9](:[0-5][0-9]){1,2}$"
        return helper.validate_param_value(
            self.name,
            value,
            str,
            regex=regex,
            exception=exception,
            valid_example='23:45',
        )


class EmailSetting(SettingValue):
    """Representation of a Email setting type."""

    def __init__(self, config: dict):
        super().__init__(config)
        self.type = list if self.multi_select else str

    def _validate_value(self, value, exception=True):
        regex = r"[^@]+@[^@]+\.[^@]+"
        return helper.validate_param_value(
            self.name,
            value,
            self.type,
            special_values=[''],
            regex=regex,
            exception=exception,
            valid_example='name@mail.com',
        )


class ObjectSetting(SettingValue):
    """Representation of an Object setting type."""

    def __init__(self, config: dict):
        super().__init__(config)
        self.type = str
        self.object_type = config.get('object_type')


class MstrObjectSetting(SettingValue):
    """Representation of an MstrObject setting type."""

    def __init__(self, config: dict):
        super().__init__(config)
        self.type = str
        self.object_type = config.get('object_type')


class DeprecatedSetting:
    """Representation of a Deprecated setting type."""

    def __init__(self, config: dict):
        self.value = None
        self.name = config.get('name')
        self.description = config.get('description')

    def _validate_value(self, value):
        msg = (
            f"Setting '{self.name}' with value '{value}' is deprecated and is read-only"
        )
        helper.exception_handler(msg=msg, exception_type=DeprecationWarning)
        return True

    def __repr__(self):
        return str(self._get_value())

    def _get_value(self):
        return self.value

    @property
    def info(self):
        return self.__dict__



"""This is the sample script to show how to manage schema facts.
This script will not work without replacing parameters with real values.
Its basic goal is to present what can be done with this module and to ease
its usage.
"""

from mstrio.connection import Connection, get_connection
from mstrio.modeling.expression import (
    ColumnReference, Expression, ExpressionFormat, FactExpression
)
from mstrio.modeling.schema.fact import Fact, list_facts
from mstrio.modeling.schema.helpers import DataType, ObjectSubType
from mstrio.modeling.schema import SchemaManagement, SchemaObjectReference, SchemaUpdateType

# For every object we want to reference using a SchemaObjectReference we need
# to provide an Object ID for. For the script to work correctly all occurrences
# of `'<object_id>'` and others with form `<some_name>` need to be replaced with
# data specific to the object used.

# Define a variable which can be later used in a script
PROJECT_NAME = $project_name  # Insert name of project here

conn: Connection = get_connection(workstationData, PROJECT_NAME)

# Define a variable which can be later used in a script
FOLDER_ID = $folder_id  # Insert folder ID here

# Example of fact data used to create new fact specified as dict
FACT_DATA = {
    'name': '<fact_name>',
    'sub_type': ObjectSubType.FACT,
    'destination_folder': FOLDER_ID,
    'data_type': {
        'type': 'float', 'precision': 8, 'scale': -2147483648
    },
    'expressions': [
        {
            'expression': {
                'tokens': [{
                    'value': 'TOT_DOLLAR_SALES + TOT_COST'
                }]
            },
            'tables': [
                {
                    'objectId': '<object_id>', 'subType': 'logical_table', 'name': '<table_name>'
                },
                {
                    'objectId': '<object_id>', 'subType': 'logical_table', 'name': '<table_name>'
                }
            ]
        }
    ]
}

# fact expression data with expression specified as tree
FACT_EXP_DATA = FactExpression(
    expression=Expression(
        tree=ColumnReference(
            column_name='<column_name>',
            object_id='<object_id>',
        )
    ),
    tables=[
        SchemaObjectReference(
            name='<table_name>',
            sub_type=ObjectSubType.LOGICAL_TABLE,
            object_id='<object_id>',
        ),
    ],
)

# DataType object used to alter fact's data_type value
DATA_TYPE_DATA = DataType(type='float', precision=6, scale=-2147483648)

# Facts management
# Get list of facts, with examples of different conditions
list_of_all_facts = list_facts(connection=conn)
list_of_limited_facts = list_facts(connection=conn, limit=10)
list_of_limited_facts_to_dict = list_facts(connection=conn, limit=10, to_dictionary=True)

# Define a variable which can be later used in a script
FACT_NAME = $fact_name  # Insert name of existing fact here

list_of_facts_by_name = list_facts(connection=conn, name=FACT_NAME)

# Define a variable which can be later used in a script
FACT_ID = $fact_id  # Insert ID of existing fact here

# Get fact by id with expressions represented as tree (default value)
fact = Fact(connection=conn, id=FACT_ID)

# Get fact by id with expressions represented as tokens
fact = Fact(connection=conn, id=FACT_ID, show_expression_as=ExpressionFormat.TOKENS)

# Get fact by name
fact_by_name = Fact(connection=conn, name=FACT_NAME)

# list of facts with expressions represented as trees
list_of_facts_as_trees = list_facts(connection=conn, show_expression_as=ExpressionFormat.TREE)

# list of facts with expressions represented as tokens
list_of_facts_as_tokens = list_facts(connection=conn, show_expression_as=ExpressionFormat.TOKENS)

# create new fact
test_fact = Fact.create(
    connection=conn,
    name=FACT_DATA['name'],
    destination_folder=FACT_DATA['destination_folder'],
    expressions=FACT_DATA['expressions'],
    data_type=FACT_DATA['data_type']
)

# Any changes to a schema objects must be followed by schema_reload
# in order to use them in reports, dashboards and so on
schema_manager = SchemaManagement(connection=conn, project_id=conn.project_id)
task = schema_manager.reload(update_types=[SchemaUpdateType.LOGICAL_SIZE])

# Add new expression to the fact
test_fact.add_expression(FACT_EXP_DATA)
first_exp_id = test_fact.expressions[0].id

# Remove first expression from the fact
test_fact.remove_expression(first_exp_id)

# Define variables which can be later used in a script
FACT_NEW_NAME = $fact_new_name  # Insert new name of edited Fact here
FACT_NEW_DESCRIPTION = $fact_new_description  # Insert new description of edited fact here

# Alter fact
test_fact.alter(name=FACT_NEW_NAME, description=FACT_NEW_DESCRIPTION)
test_fact.alter(data_type=DATA_TYPE_DATA)

# Delete newly created fact
test_fact.delete(force=True)

# List all tables for fact
tables_all = fact.get_tables()

# List tables for fact expression
exp_id = fact.expressions[0].id
exp_obj = fact.expressions[0]
tables_exp_id = fact.get_tables(expression=exp_id)
tables_exp_obj = fact.get_tables(expression=exp_obj)

# List fact properties
properties = fact.list_properties()



import functools
import logging
import sys
from enum import Enum, auto
from itertools import filterfalse
from textwrap import dedent

from packaging.version import Version

from mstrio import __version__, config
from mstrio.helpers import NotSupportedError

logger = logging.getLogger(__name__)
log_handler = logging.StreamHandler(sys.stderr)
log_handler.setLevel(logging.INFO)
log_handler.setFormatter(logging.Formatter('%(message)s'))
logger.addHandler(log_handler)

current_version = Version(__version__)


class WipLevels(Enum):
    SILENT = auto()
    PREVIEW = auto()
    INFO = auto()
    WARNING = auto()
    ERROR = auto()


def _wiplevel_error(level):
    raise ValueError(f'Unrecognised WipLevel {level}.')


message_templates = {
    WipLevels.PREVIEW: (
        "In a current version, {} is available as "
        "a Functionality Preview. It is subject to change until it is released "
        "as Generally Available."
    ),
    WipLevels.INFO: (
        "This {} functionality is a work-in-progress. It may change in future updates."
    ),
    WipLevels.WARNING: (
        "This {} functionality is a work-in-progress. Use it only if you understand"
        " the underlying code. It may change in future updates."
    ),
    WipLevels.ERROR: (
        "This {} functionality is a work-in-progress. It has been deemed unsafe to use"
        " and is currently blocked."
    ),
}
release_info_template = " It is currently scheduled for {} release."
docstring_prefix_template = dedent(
    """\
    ------------------ WORK IN PROGRESS ------------------

    {}

    ------------------------------------------------------

"""
)


def _get_message_template(level: WipLevels = WipLevels.WARNING):
    if level in message_templates:
        return message_templates[level]
    else:
        return message_templates[WipLevels.INFO]


def _construct_message(
    name: str | None = None,
    target_release: Version | str | None = None,
    level: WipLevels = WipLevels.WARNING,
):
    template = _get_message_template(level)
    if name is None:
        message = template.format("")
    else:
        message = template.format(f"({name}) ")
    if target_release is not None:
        message += release_info_template.format(target_release)
    return message


def _emit(
    name: str | None = None,
    target_release: Version | str | None = None,
    level: WipLevels = WipLevels.WARNING,
    message: str | None = None,
):
    if level == WipLevels.SILENT or not config.wip_warnings_enabled:
        return None

    if message is None:
        message = _construct_message(name, target_release, level)

    if level == WipLevels.INFO or level == WipLevels.PREVIEW:
        logger.info(message)
    elif level == WipLevels.WARNING:
        logger.warning(message)
    elif level == WipLevels.ERROR:
        raise NotSupportedError(message)
    else:
        _wiplevel_error(level)


def wip(
    target_release: Version | str | None = None,
    level: WipLevels = WipLevels.WARNING,
    message: str | None = None,
    prefix_doc: bool | str = True,
    mark_attr: bool = True,
):
    """Work-In-Progress wrapper

    Note:
        This is a decorator generator, which accepts arguments and returns the
        actual decorator, so even when not providing arguments and letting it
        choose the defaults it should be used like `@wip()`, not plain `@wip`.

    Args:
        target_release: The target release when the functionality is scheduled
            to be production-ready.
        level: The severity level of the warning emitted when the functionality
            is loaded or used.
            SILENT: no warning will be emitted
            PREVIEW: a warning will be dispatched
            INFO: a disclaimer will be printed to stderr
            WARNING: a warning will be dispatched
            ERROR: an error will be raised
        message: A custom message to replace the one standard to the given level
        prefix_doc: Whether to prefix the docstring with a WIP warning.
            Accepts bool values or a custom string to replace standard prefix.
        mark_attr: Marks the wrapped function via adding the `_wip` attribute,
            which is checked by `is_wip`.
    """
    if target_release is not None:
        if not isinstance(target_release, Version):
            target_release = Version(target_release)
        if target_release <= current_version:
            raise ValueError(
                "WIP wrapper called with target_release equal to"
                " or lower than current release."
            )
    if level not in WipLevels:
        _wiplevel_error(level)

    def wrap_func(f):
        @functools.wraps(f)
        def wrapped(*args, **kwargs):
            _emit(wrapped.__name__, target_release, level, message)
            return f(*args, **kwargs)

        if mark_attr:
            wrapped._wip = True  # This makes it trivial to programmatically check.
        if prefix_doc:
            if not wrapped.__doc__:
                wrapped.__doc__ = ""
            if isinstance(prefix_doc, str):
                wrapped.__doc__ = prefix_doc + wrapped.__doc__
            else:
                if message is not None:
                    docstring_prefix = docstring_prefix_template.format(message)
                else:
                    msg = _construct_message(target_release=target_release, level=level)
                    docstring_prefix = docstring_prefix_template.format(msg)
                wrapped.__doc__ = docstring_prefix + wrapped.__doc__
        return wrapped

    return wrap_func


def module_wip(
    module_globals: dict,
    target_release: Version | str | None = None,
    level: WipLevels = WipLevels.WARNING,
    message: str | None = None,
    prefix_doc: bool | str = True,
    mark_attr: bool = True,
):
    """Emit the WIP warning/error/info when the module is loaded."""
    if mark_attr:
        module_globals["_wip"] = True
    if prefix_doc:
        if not module_globals["__doc__"]:
            module_globals["__doc__"] = ""
        if isinstance(prefix_doc, str):
            module_globals["__doc__"] = prefix_doc + module_globals["__doc__"]
        else:
            if message is not None:
                docstring_prefix = docstring_prefix_template.format(message)
            else:
                msg = _construct_message(target_release=target_release, level=level)
                docstring_prefix = docstring_prefix_template.format(msg)
            module_globals["__doc__"] = docstring_prefix + module_globals["__doc__"]
    _emit(f"module {module_globals['__name__']}", target_release, level, message)


def pause_wip_warnings():
    previous_setting = config.wip_warnings_enabled
    config.wip_warnings_enabled = False
    return previous_setting


def resume_wip_warnings(previous_setting: bool):
    config.wip_warnings_enabled = previous_setting


def is_wip(obj) -> bool:
    return getattr(obj, "_wip", False)


def is_module_wip(module_globals: dict) -> bool:
    return module_globals.get("_wip", False)


def filter_wip(iterable, return_wip: bool = False):
    """Return a generator containing only non-WIP elements
        (or only WIP elements, depending on `return_wip` parameter.).

    Args:
        iterable: Any iterable containing elements that may or may not be WIP.
        return_wip: Whether to return only WIP results or the opposite.
            Default `False` (return only non-WIP).
    """
    if return_wip:
        return filter(is_wip, iterable)
    return filterfalse(is_wip, iterable)


def filter_wip_dict(dict: dict, return_wip: bool = False) -> dict:
    """Return a dict containing only non-WIP elements
        (or only WIP elements, depending on `return_wip` parameter.).

    Args:
        dict: Any dict containing elements that may or may not be WIP.
        return_wip: Whether to return only WIP results or the opposite.
            Default `False` (return only non-WIP).
    """
    if return_wip:
        return {key: val for key, val in dict.items() if is_wip(val)}
    return {key: val for key, val in dict.items() if not is_wip(val)}
